/*
 * kernel/file.c
 * Copyright (C) 1996, Onno Hovers, All rights reserved
 *
 * functions that act on file objects
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; see the file COPYING.LIB. If
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 * Cambridge, MA 02139, USA.  
 */

#include <stddef.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include <win32/defines.h>
#include <win32/error.h>
#include <win32/kernel.h>

#include "wfc.h"
#include "itime.h"
#include "teb.h"
#include "imalloc.h"
#include "object.h"
#include "file.h"
#include "fs.h"
#include "error.h"

FILE_CLASS __FileClass =
{ { { FILE_GENERIC_READ, FILE_GENERIC_WRITE, 
      FILE_GENERIC_EXECUTE, FILE_ALL_ACCESS },
       __FileClose, __FileClose,  __FileIsType, 
       __ObjectGetInfoHandle, __ObjectSetInfoHandle, __ObjectDuplicate, 
       __ObjectWait }, 
  { __FileRead, __FileWrite, __FileSeek, __FileSetEOF, __FileGetFileInfo, 
    __FileGetSize, __FileGetFileType, __FileGetTime, __FileSetTime, 
    __FileFlush, __FileIoControl, __FileLock, __FileUnlock } };

BOOL  __FileInit(VOID)
{
   PFILE_OBJECT pfile;
   
   /* init stdin */
   pfile=imalloc(sizeof(FILE_OBJECT));
   if(!pfile)
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY); 
   pfile->Kernel.Magic=MAGIC_FILE;
   pfile->Kernel.Class=(PKERNEL_CLASS) &__FileClass;
   pfile->Kernel.HandleCount=1;
   pfile->File.FileDesc=0; 
   pfile->File.Pointer=0;
   __HStdIn =__HandleAlloc((PKERNEL_OBJECT) pfile, FILE_GENERIC_READ );  
   if(!__HStdIn)
   { 
      ifree(pfile); 
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY);
   };
   
   /* init stdout */
   pfile=imalloc(sizeof(FILE_OBJECT));
   if(!pfile)
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY); 
   pfile->Kernel.Magic=MAGIC_FILE;
   pfile->Kernel.Class=(PKERNEL_CLASS) &__FileClass;
   pfile->Kernel.HandleCount=1;
   pfile->File.FileDesc=1; 
   pfile->File.Pointer=0;
   __HStdOut=__HandleAlloc((PKERNEL_OBJECT) pfile, FILE_GENERIC_WRITE );  
   if(!__HStdOut)
   { 
      ifree(pfile); 
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY);
   };
   
   /* init stderr */
   pfile=imalloc(sizeof(FILE_OBJECT));
   if(!pfile)
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY); 
   pfile->Kernel.Magic=MAGIC_FILE;
   pfile->Kernel.Class=(PKERNEL_CLASS) &__FileClass;
   pfile->Kernel.HandleCount=1;
   pfile->File.FileDesc=2; 
   pfile->File.Pointer=0;
   __HStdErr=__HandleAlloc((PKERNEL_OBJECT) pfile, FILE_GENERIC_WRITE );  
   if(!(__HStdErr))
   { 
      ifree(pfile); 
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY);
   };

   return TRUE;
}

BOOL  __FileOpen(PHANDLE phnd, LPCSTR filename, DWORD access, DWORD share, 
		LPSECURITY_ATTRIBUTES psa, DWORD create, DWORD flagsattr, 
		HANDLE htemp)
{  
   PFILE_OBJECT pfile;
   LPSTR pcertain;
   CHAR  fullname[PATH_MAX];
   CHAR  unixname[PATH_MAX];
   DWORD findflags;
   int   fd;
   int   oflags;
   int   mode;
   
   *phnd=INVALID_HANDLE_VALUE;
   
   oflags = __FsAccessFlags(access) | __FsCreationFlags(create);
   dprintf("oflags = 0%o\n", oflags);
   mode   = __FsAttribToMode(flagsattr);
   dprintf("mode  = 0%o\n", mode);
   if(!__FsGetNameA(filename, fullname, unixname, &pcertain, PATH_MAX))
      return FALSE;
   
   fd=open(unixname, oflags, mode);
   if(fd==-1)
   {
      findflags=__FsFindNameFlags(create);
      
      if(!__FsFindNameA(fullname, unixname, pcertain, PATH_MAX, findflags))
         return FALSE;
      
      fd=open(unixname, oflags, mode);
      if(fd==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));            
   }
   
   /* we have a file descriptor, now allocate a handle and object */
   pfile=imalloc(sizeof(FILE_OBJECT));
   if(!pfile)
      { close(fd); return __ErrorReturnFalse(ERROR_OUTOFMEMORY); }
   pfile->Kernel.Magic=MAGIC_FILE;
   pfile->Kernel.Class=(PKERNEL_CLASS) &__FileClass;
   pfile->Kernel.HandleCount=1;
   pfile->File.FileDesc=fd; 
   pfile->File.Pointer=0;
   *phnd=__HandleAlloc((PKERNEL_OBJECT) pfile, access);  
   if(!(*phnd))
   { 
      ifree(pfile); close(fd); 
      *phnd=INVALID_HANDLE_VALUE;
      return __ErrorReturnFalse(ERROR_OUTOFMEMORY);
   };
   return TRUE;
}

BOOL  __FileIsType(PKERNEL_OBJECT pfile, ULONG typeid)
{
   if(pfile->Kernel.Magic!=MAGIC_FILE ) 
      return __ErrorReturnFalse(ERROR_INVALID_HANDLE );
   if(typeid==MAGIC_FILE)
      return TRUE;
   else if(typeid==MAGIC_OBJECT)
      return TRUE;
   else
      return __ErrorReturnFalse(ERROR_INVALID_HANDLE);
}

BOOL  __FileClose(PKERNEL_OBJECT pfile)
{
   /* XXX: should be InterlockedDecrement */ 
   if(--(pfile->Kernel.HandleCount))
      return TRUE;
   if(close( ((PFILE_OBJECT) pfile)->File.FileDesc)==-1)
      return __ErrorReturnFalse(__ErrorFromErrno(errno));
   ifree(pfile);
   return TRUE;
}

BOOL  __FileRead(PFILE_OBJECT pfile, LPVOID pbuffer, DWORD toread,
               LPDWORD pread, LPOVERLAPPED poverlapped,
               LPOVERLAPPED_COMPLETION_ROUTINE polcr)
{
   int retval;

   if((poverlapped)||(polcr))
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);

   retval=read(pfile->File.FileDesc,pbuffer,toread);

   if( retval == -1 )
      return __ErrorReturnFalse(__ErrorFromErrno(errno));

   if(pread)
      *pread=retval;

   return TRUE;
}

BOOL  __FileWrite(PFILE_OBJECT pfile, LPCVOID pbuffer, DWORD towrite,
                LPDWORD pwritten, LPOVERLAPPED poverlapped,
                LPOVERLAPPED_COMPLETION_ROUTINE polcr)
{
   int retval;
   
   if((poverlapped)||(polcr))
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);

   retval=write(pfile->File.FileDesc,pbuffer,towrite);

   if( retval == -1 )
      return __ErrorReturnFalse(__ErrorFromErrno(errno));  

   if(pwritten)
      *pwritten=retval;

   return TRUE;
}

#define NOSEEK  (ULONG) -1

BOOL __FileSeek(PFILE_OBJECT pobj, PULONG pseeklow, PULONG pseekhigh, ULONG whence)
{ 
#ifdef USE_LLSEEK
   ULONG     retval;
   ULONG     seekhigh;
   long long llfileptr;

   if(!pseekhigh)
      seekhigh=0;
   else
      seekhigh=*pseekhigh;

   retval=_llseek(pobj->File.FileDesc, seekhigh, *pseeklow, (loff_t) llfileptr, whence);

   if(retval==-1)
   {
      if(*pseekhigh)
         *pseekhigh=NOSEEK;
      *pseeklow=NOSEEK;
      return __ErrorReturnFalse(__ErrorFromErrno);
   }
   
   if(pseekhigh)  
      *pseekhigh=llfileptr >> 32;
   *pseeklow= llfileptr &  0xFFFFFFFF;
   
   return retval;
#else

   if(pseekhigh)
   {     
      if(*pseekhigh)
      {
          *pseekhigh=NOSEEK;
          *pseeklow =NOSEEK;
          return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);         
      }     
   }  
   
   *pseeklow=lseek(pobj->File.FileDesc,*pseeklow,whence);
   
   if(*pseeklow==-1)
   {
      if(pseekhigh)
         *pseekhigh=NOSEEK;
      *pseeklow=NOSEEK;
      return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }

   return TRUE;

#endif
}

BOOL  __FileSetEOF( PFILE_OBJECT pfile )
{ 
   ULONG fpos;
   
   fpos=lseek(pfile->File.FileDesc,0,SEEK_CUR);
   if(fpos==-1)
      return __ErrorReturnFalse(__ErrorFromErrno(errno)); 
   if(ftruncate(pfile->File.FileDesc,fpos)!=0)
      return __ErrorReturnFalse(__ErrorFromErrno(errno)); 
   return TRUE;
}

BOOL __FileGetFileInfo(PFILE_OBJECT pfile, PBY_HANDLE_FILE_INFORMATION pinfo)
{
   struct stat mystat;  

   if(fstat(pfile->File.FileDesc, &mystat)==-1);
      return __ErrorReturnFalse(__ErrorFromErrno(errno));

   /* translate the file attributes */
   pinfo->dwFileAttributes = __FsModeToAttrib(mystat.st_mode);
   
   /* translate the file times */
   pinfo->ftCreationTime  =__TimeUnixToFile(mystat.st_ctime);
   pinfo->ftLastWriteTime =__TimeUnixToFile(mystat.st_mtime);
   pinfo->ftLastAccessTime=__TimeUnixToFile(mystat.st_atime);

   pinfo->nFileSizeLow   = mystat.st_size;
   pinfo->nNumberOfLinks = mystat.st_nlink;
   pinfo->nFileIndexLow  = mystat.st_ino;

   /* zero out currently unused fields */   
   pinfo->dwVolumeSerialNumber = 0;
   pinfo->nFileSizeHigh = 0;
   pinfo->nFileIndexHigh = 0;

   return TRUE;
}

BOOL  __FileGetFileType( PFILE_OBJECT pfile, PULONG pfiletype )
{
#ifdef NOTYET
    *pfiletype=FILE_TYPE_DISK;
    return TRUE;    
#else    
    struct stat mystat;
    
    mystat.st_mode=0;
    fstat(pfile->File.FileDesc, &mystat);
    switch(mystat.st_mode & S_IFMT)
    {
       
       case S_IFCHR:
          *pfiletype=FILE_TYPE_CHAR;
          return TRUE;
       case S_IFIFO:
       case S_IFSOCK:
          *pfiletype=FILE_TYPE_PIPE;
          return TRUE;
       case S_IFREG:
       case S_IFBLK:
       default:
          *pfiletype=FILE_TYPE_DISK;    
          return TRUE;
    }
#endif
}

BOOL  __FileGetSize( PFILE_OBJECT pfile, PULONGLONG psize )
{
   struct stat mystat;
   
   *psize=0;
   if(fstat(pfile->File.FileDesc, &mystat)!=0)
       return __ErrorReturnFalse(__ErrorFromErrno(errno));
   *psize=mystat.st_size;
   return TRUE;
}

BOOL __FileGetTime( PFILE_OBJECT pfile, PFILETIME pcrtime, 
                  PFILETIME patime, PFILETIME pmtime)
{
   struct stat mystat;  

   if(fstat(pfile->File.FileDesc, &mystat)==-1);
      return __ErrorReturnFalse(__ErrorFromErrno(errno));

   /* translate the file times */
   if(pcrtime)
      *pcrtime=__TimeUnixToFile(mystat.st_ctime);
   if(patime)
      *patime=__TimeUnixToFile(mystat.st_mtime);
   if(pmtime)
      *pmtime=__TimeUnixToFile(mystat.st_atime);

   return TRUE;
}

BOOL __FileSetTime( PFILE_OBJECT pfile, CONST FILETIME *pctime,
                  CONST FILETIME *patime, CONST FILETIME *pmtime)
{
#ifdef NOTYET
   /* We cannot change ctime and have to give the entire filename
    * of the file as argument!
    * How I just hate UNIXen!!
    */
   ULONG        atime, mtime;
   struct stat  mystat;
   
   if((!patime)&&(!pmtime))
      return TRUE;
      
   if((!patime)||(!pmtime))
   {       
       if(fstat(pfile->File.FileDesc,&mystat)==-1)
          return __ErrorReturnFalse(__ErrorFromErrno(errno));        
   }
   
   if(patime)
      atime=__TimeFileToUnix(patime);
   else
      atime=mystat.st_atime;
      
   if(pmtime)
      mtime=__TimeFileToUnix(pmtime);
   else
      mtime=mystat.st_mtime;
          
   utime(*****);
#else
   return __ErrorReturnFalse(ERROR_CALL_NOT_IMPLEMENTED);   
#endif            
}

BOOL __FileFlush( PFILE_OBJECT pfile )
{
   if(fsync(pfile->File.FileDesc)==-1)
      return __ErrorReturnFalse(__ErrorFromErrno(errno));
   return TRUE;
}

BOOL __FileIoControl( PFILE_OBJECT pfile, ULONG access, 
                    ULONG iocode, PVOID inbuf, 
                    ULONG insize, PVOID outbuf, ULONG outsize, 
                    PULONG poutsize, POVERLAPPED pol)
{
   return __ErrorReturnFalse(ERROR_INVALID_CATEGORY);
}

BOOL __FileLock( PFILE_OBJECT pfile, ULONG flags, ULONG reserved,
               ULONGLONG offset, ULONGLONG size, POVERLAPPED pol)
{
   return __ErrorReturnFalse(ERROR_CALL_NOT_IMPLEMENTED);
}

BOOL __FileUnlock( PFILE_OBJECT pfile, ULONG reserved,
                 ULONGLONG offset, ULONGLONG size, POVERLAPPED pol)
{
   return __ErrorReturnFalse(ERROR_CALL_NOT_IMPLEMENTED);
}


/*********************************************************************
*                      ReadFile  --   KERNEL32                       *
*                                                                    *
* Read from a file object                                            *
*                                                                    *
*********************************************************************/

BOOL WINAPI ReadFile(HANDLE hfile, LPVOID pbuf, DWORD toread,
                     LPDWORD pread, LPOVERLAPPED pol)
{
    PFILE_OBJECT pfile;
    
    aprintf("ReadFile( 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n",
            (ULONG) hfile, (ULONG) pbuf, toread, (ULONG) pread, 
            (ULONG) pol);
            
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_GENERIC_READ, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL(pfile, Read, 
                             (pfile, pbuf, toread, pread, pol, NULL));
}

/*********************************************************************
*                      ReadFileEx  --   KERNEL32                     *
*                                                                    *
* Read from a file object asynchroneously                            *
*                                                                    *
*********************************************************************/

BOOL WINAPI ReadFileEx(HANDLE hfile, LPVOID pbuf,
                       DWORD toread, LPOVERLAPPED pol,
                       LPOVERLAPPED_COMPLETION_ROUTINE polcr)
{
    PFILE_OBJECT pfile;
    
    aprintf("ReadFileEx( 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n",
            (ULONG) hfile, (ULONG) pbuf, toread, (ULONG) pol, 
            (ULONG) polcr);

    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_GENERIC_READ, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL(pfile, Read, 
                             (pfile, pbuf, toread, NULL, pol, polcr));
}

/*********************************************************************
*                      WriteFile  --  KERNEL32                       *
*                                                                    *
* Write to a file object                                             *
*                                                                    *
*********************************************************************/

BOOL WINAPI WriteFile(HANDLE hfile, LPCVOID pbuf, DWORD towrite,
                      LPDWORD pwritten, LPOVERLAPPED pol)
{
    PFILE_OBJECT pfile;
    
    aprintf("WriteFile( 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n",
            (ULONG) hfile, (ULONG) pbuf, towrite, (ULONG) pwritten, 
            (ULONG) pol);
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_GENERIC_WRITE, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL(pfile, Write, 
                             (pfile, pbuf, towrite, pwritten, pol, NULL));
}

/*********************************************************************
*                      WriteFileEx  --  KERNEL32                     *
*                                                                    *
* Write to a file object asynchroneously                             *
*                                                                    *
*********************************************************************/

BOOL WINAPI WriteFileEx(HANDLE hfile, LPCVOID pbuf,
                        DWORD towrite, LPOVERLAPPED pol,
                        LPOVERLAPPED_COMPLETION_ROUTINE polcr)
{
    PFILE_OBJECT pfile;
    
    aprintf("WriteFileEx( 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n",
            (ULONG) hfile, (ULONG) pbuf, towrite, (ULONG) pol, 
            (ULONG) polcr);
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_GENERIC_WRITE, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL(pfile, Write, 
                             (pfile, pbuf, towrite, NULL, pol, polcr));
}

/********************************************************************
*                  SetFilePointer  --  KERNEL32                     *
*                                                                   *
*  Set the pointer of a file                                        *
*                                                                   *
********************************************************************/

DWORD WINAPI SetFilePointer(HANDLE hfile, LONG seek, 
                            LPLONG pseekhi, DWORD whence)
{
    PFILE_OBJECT pfile;
    
    aprintf("SetFilePointer( 0x%lX, %ld, 0x%lX, %ld )\n",
            (ULONG) hfile, seek,  (ULONG) pseekhi, whence );            
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, 0, 
                                      TYPEID_FILE);
    if(!pfile)
    {
       if(pseekhi)
          *pseekhi=-1;
       return (DWORD) -1;
    }
   
    PFILE_OBJECT_CALL( pfile, Seek, (pfile, &seek, pseekhi, whence) );
    
    return seek;
}

/********************************************************************
*                     SetEndOfFile  --  KERNEL32                    *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI SetEndOfFile(HANDLE hfile)
{
    PFILE_OBJECT pfile;

    aprintf("SetEndOfFile( 0x%lX )\n", (ULONG) hfile );
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_GENERIC_WRITE, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL( pfile, SetEOF, (pfile) );
}

/********************************************************************
*             GetFileInformationByHandle  --  KERNEL32              *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI GetFileInformationByHandle(HANDLE hfile, 
                                       LPBY_HANDLE_FILE_INFORMATION pinfo)
{
    PFILE_OBJECT pfile;
    
    aprintf("GetFileInformationByHandle( 0x%lX, 0x%lX )\n", 
            (ULONG) hfile, (ULONG) pinfo );
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_READ_ATTRIBUTES,
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
    
    return PFILE_OBJECT_CALL( pfile, GetFileInfo, (pfile, pinfo) );
}


/********************************************************************
*                     GetFileType  --  KERNEL32                     *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

DWORD WINAPI GetFileType(HANDLE hfile)
{
    PFILE_OBJECT pfile;
    ULONG        type;

    aprintf("GetFileType( 0x%lX )\n", (ULONG) hfile );    
            
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, 0, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
        
    PFILE_OBJECT_CALL( pfile, GetFileType, (pfile, &type) );
    return type;
}


/********************************************************************
*                     GetFileSize  --  KERNEL32                     *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

DWORD WINAPI GetFileSize(HANDLE hfile, PULONG phisize)
{
   PFILE_OBJECT	pfile;
   ULONGLONG	size;

   aprintf("GetFileSize( 0x%lX, 0x%lX )\n", (ULONG) hfile, (ULONG) phisize );

   pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_READ_ATTRIBUTES, 
                                     TYPEID_FILE);
   if(pfile)
   {  
      PFILE_OBJECT_CALL( pfile, GetSize, (pfile, &size) );
      if(phisize)
         *phisize=HILONG(size);
    
      return LOLONG(size);
   }
   else
      return 0;          
}

/********************************************************************
*                     GetFileTime  --  KERNEL32                     *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI GetFileTime(HANDLE hfile, PFILETIME pctime, 
                        PFILETIME patime, PFILETIME pmtime)
{
    PFILE_OBJECT pfile;
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_READ_ATTRIBUTES, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;
   
    return PFILE_OBJECT_CALL( pfile, GetTime, 
                             (pfile, pctime, patime, pmtime) );

}

/********************************************************************
*                     SetFileTime  --  KERNEL32                     *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI SetFileTime(HANDLE hfile, CONST FILETIME *pctime, 
                        CONST FILETIME *patime, CONST FILETIME *pmtime)
{
    PFILE_OBJECT pfile;      

    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, FILE_WRITE_ATTRIBUTES, 
                                      TYPEID_FILE);
    if(!pfile)
       return FALSE;

    return  PFILE_OBJECT_CALL(pfile, SetTime, 
                              (pfile, pctime, patime, pmtime) );

}

/********************************************************************
*                 FlushFileBuffers  --  KERNEL32                    *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI FlushFileBuffers(HANDLE hfile)
{
    PFILE_OBJECT pfile;
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, 0, TYPEID_FILE);
    if(!pfile)
        return FALSE;
   
    return PFILE_OBJECT_CALL( pfile, Flush, (pfile) );
   
}

/********************************************************************
*                    DeviceIoControl  --  KERNEL32                  *
*                                                                   *
*                                                                   *
*                                                                   *
********************************************************************/

BOOL WINAPI DeviceIoControl(HANDLE hfile, ULONG iocode, PVOID inbuf, 
                            ULONG insize, PVOID outbuf, ULONG outsize, 
                            PULONG poutsize, POVERLAPPED pol)
{
    PHANDLE_DATA phand;         

    phand=__HandleGet(hfile);
    
    if( (!phand) ||
        (!PKERNEL_OBJECT_CALL( phand->Object, IsType, 
                              (phand->Object, TYPEID_FILE))) )
       return __ErrorReturnFalse(ERROR_INVALID_HANDLE);
                 
    return PFILE_OBJECT_CALL( phand->Object, IoControl, 
                              ((PFILE_OBJECT)(phand->Object), 
                               phand->Access, iocode, 
                               inbuf, insize, outbuf, outsize, 
                               poutsize, pol) );
}

/*********************************************************************
*                      LockFile  --   KERNEL32                       *
*                                                                    *
* Lock a range from a file object                                    *
*                                                                    *
*********************************************************************/

BOOL WINAPI LockFile(HANDLE hfile, ULONG offsetlo, ULONG offsethi,
                     ULONG sizelo, ULONG sizehi)
{
    PFILE_OBJECT pfile;
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, SYNCHRONIZE, TYPEID_FILE);
    if(!pfile)
       return FALSE;

    return PFILE_OBJECT_CALL(pfile, LockFile, 
                                (pfile, 0, 0, 
                                 MAKELONGLONG(offsetlo, offsethi),
                                 MAKELONGLONG(sizelo,sizehi), NULL));
}

/*********************************************************************
*                      LockFileEx  --   KERNEL32                     *
*                                                                    *
* Lock a range from a file object                                    *
*                                                                    *
*********************************************************************/

BOOL WINAPI LockFileEx(HANDLE hfile, DWORD flags, DWORD reserved,
                       DWORD sizelo, DWORD sizehi, LPOVERLAPPED pol)
{
    PFILE_OBJECT pfile;
    
    
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, SYNCHRONIZE, TYPEID_FILE);
    if(!pfile)
       return FALSE;

    return PFILE_OBJECT_CALL(pfile, LockFile, 
                                 (pfile, flags, reserved, 0,
                                  MAKELONGLONG(sizelo, sizehi), pol));
}

/*********************************************************************
*                     UnlockFile  --  KERNEL32                       *
*                                                                    *
* Unlock a range from a file object                                  *
*                                                                    *
*********************************************************************/

BOOL WINAPI UnlockFile(HANDLE hfile, ULONG offsetlo, ULONG offsethi,
                       ULONG sizelo, ULONG sizehi)
{
    PFILE_OBJECT pfile;

    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, SYNCHRONIZE, TYPEID_FILE);
    if(!pfile)
       return FALSE;

    return PFILE_OBJECT_CALL(pfile, UnlockFile,
                                (pfile, 0, 
                                MAKELONGLONG(offsetlo,offsethi),
                                MAKELONGLONG(sizelo,sizehi), NULL));
}

/*********************************************************************
*                     UnlockFileEx  --  KERNEL32                     *
*                                                                    *
* Unlock a range from a file object                                  *
*                                                                    *
*********************************************************************/

BOOL WINAPI UnlockFileEx(HANDLE hfile, DWORD reserved,
                         DWORD sizelo, DWORD sizehi, LPOVERLAPPED pol)                      
{
    PFILE_OBJECT pfile;

        
    pfile=(PFILE_OBJECT) __HandleGetObject(hfile, SYNCHRONIZE, TYPEID_FILE);
    if(!pfile)
       return FALSE;

    return PFILE_OBJECT_CALL(pfile, UnlockFile, 
                             (pfile, reserved, 0, 
                             MAKELONGLONG(sizelo,sizehi), pol));    
}

/*********************************************************************
*                     GetStdHandle  --  KERNEL32                     *
*                                                                    *
* Get one of the standard handles                                    *
*                                                                    *
*********************************************************************/
                   
HANDLE WINAPI GetStdHandle(DWORD what)
{
   dprintf("GetStdHandle called\n");
   switch(what)
   {
      case STD_INPUT_HANDLE:
         return __HStdIn;
      case STD_OUTPUT_HANDLE:
         return __HStdOut;
      case STD_ERROR_HANDLE:
         return __HStdErr;
      default:
         SetLastError(ERROR_INVALID_PARAMETER);
         return (HANDLE) NULL;
   }
}
                    