/*
 *  kernel/fs.c
 *  Copyright (C) 1996, Onno Hovers, All rights reserved
 *
 *  Win32 filesystem functions
 */

/* unix includes */
#include<assert.h>
#include<unistd.h>
#include<dirent.h>
#include<string.h>
#include<ctype.h>
#include<sys/stat.h>
#include<sys/fcntl.h>

/* Win32 includes */
#include<win32/defines.h>
#include<win32/error.h>
#include<win32/kernel.h>

#include "file.h"
#include "rtlmem.h"
#include "wfc.h"
#include "error.h"

typedef struct __DRIVE
{
   LPSTR      UnixRoot;
   CHAR       VolumeLabel[12];
   ULONG      SerialNumber;
   ULONG      DriveType;
} DRIVE;

DRIVE __Drive[26];

#define FINDNAME_DIR        1
#define FINDNAME_FILE       2
#define FINDNAME_EXIST      4
#define FINDNAME_NEW        8 

int __FsBufferSize=4096;

VOID __FsChangeSeparator(LPSTR pathname, CHAR sep)
{
   LPSTR tempstr;
   
   /* now change the separator to what we want */
   tempstr=pathname;
   while(*tempstr)
   {
      if((*tempstr=='\\')||(*tempstr=='/'))
         *tempstr=sep;
      tempstr++;      
   }
}


BOOL  __FsInit()   
{
   LPSTR curdir;
   LPSTR driveinf;
   LPSTR temp;
   CHAR  lastavail;
   CHAR  rootdrive;
   CHAR  envname[]="WF_DRIVE_A";

   lastavail=0;
   rootdrive=0;
   
   /* initialize drives */
   for(envname[9]='A'; envname[9]<='Z'; envname[9]++)
   {
      driveinf=getenv(envname);
      if(driveinf)
      {
         /* is this root ?? */         
         temp=driveinf;       
         while(*temp=='/') 
            temp++;
         if(*temp=='\0')
            rootdrive=envname[9];
            
         __Drive[envname[9]-'A'].UnixRoot=driveinf;
      }           
      else
         lastavail=envname[9];     
   }
   
   /* we need a drive mapped to '/' */
   if(!rootdrive)
   {
      if(!lastavail)
      {
         fprintf(stderr,
                 "Warning: no drive for root directory, defaulting to Z:\n");
         lastavail='Z';
      }        
      __Drive[lastavail-'A'].UnixRoot="/";
      rootdrive=lastavail;      
   }   
   getcwd(__UnixCurrentDirectory,MAX_PATH);

   __CurrentDirectoryA[0]=rootdrive; 
   __CurrentDirectoryA[1]=':';
   istrcat(__CurrentDirectoryA,__UnixCurrentDirectory, MAX_PATH);

   curdir=__CurrentDirectoryA;
   while(*curdir) 
   {
      if(*curdir=='/')
         *curdir='\\';
      curdir++;
   }   
   return TRUE;
}

/********************************************************************
*                   __FsFindNameA                                   *
*                                                                   *
* This is a last resort function:                                   *
* if someone tries to open                                          *
* "C:\USR\LOCAL\DOC\WFC\README"  and "C:\" is mapped to "/",        *
* __FsGetNameA translates this to:                                  *
* "/USR/LOCAL/DOS/WFC/README"                                       *
* opening this file will not work, due to unix case sensitivity     *
* so we try to find the right name the hard way.                    *
*                                                                   *
********************************************************************/

BOOL __FsFindNameA(LPSTR fullname, LPSTR unixname, LPSTR pcertain, 
                     INT count, INT flags)
{     
   struct dirent *mydent;
   DIR           *mydir;            
   LPSTR         puncertain;
   INT           namelen; 
      
   fullname=fullname+3;  /* skip 'X:\' */
       
   while(1)
   {
       /* skip double '//' */
       while(pcertain[1]=='/')
          pcertain++;
       
      /* open directory until pcertain */      
      *pcertain = '\0';

      if(unixname[0]=='\0')      
         mydir = opendir("/");     
      else           
         mydir = opendir(unixname);        
      *pcertain='/';
      
      pcertain++;

      /* get the next part of the path */
      puncertain = strchr(pcertain,'/');         
            
      /* if this is not the last part of the path */
      if(puncertain)
      {
         namelen = puncertain-pcertain;
         *puncertain='\0';              
         do
         {                                          
            mydent=readdir(mydir);
            if(!mydent)
            {      
               /* close directory */
               closedir(mydir);
               
               __ErrorReturnFalse(ERROR_PATH_NOT_FOUND);
            } 
            dprintf("trying: %s\n",mydent->d_name); 
         }     
         while(strcasecmp(mydent->d_name, pcertain));
         *puncertain='/';
         
         strncpy(pcertain,mydent->d_name,namelen);
         strncpy(fullname,mydent->d_name,namelen);
         fullname=fullname+namelen+1;
         pcertain=puncertain;

         closedir(mydir);              
      }      
      else
      { 
         namelen = strlen(pcertain);
         
         do
         {                          
            mydent=readdir(mydir);
            if(!mydent)
            {      
               /* close directory */
               closedir(mydir);           
                              
               if((flags&FINDNAME_NEW)==0)
               {
                  if(flags&FINDNAME_DIR)
                     return __ErrorReturnFalse(ERROR_PATH_NOT_FOUND);
                  else
                     return __ErrorReturnFalse(ERROR_FILE_NOT_FOUND);
               }
               pcertain = NULL;
               return TRUE;
            }
            dprintf(" trying: %s\n",mydent->d_name);            
         }      
         while(strcasecmp(mydent->d_name, pcertain));
         
         strncpy(pcertain,mydent->d_name,namelen);
         strncpy(fullname,mydent->d_name,namelen);                  
         closedir(mydir);
         
         /* but what if we do not want an existing name? */
         if((flags&FINDNAME_EXIST)==0)
         { 
            if(flags&FINDNAME_DIR)
               return __ErrorReturnFalse(ERROR_CANNOT_MAKE);  
            else
               return __ErrorReturnFalse(ERROR_FILE_EXISTS);
         }
         return TRUE;
      }       
   }
   return TRUE; 
}

/*******************************************************************
*                       __FsGetNameA                               *
*                                                                  *
* This function makes a unix name from the windows name, but does  *
* not check if this name really exist. This way we only have to    *
* browse through the directories when this fails.                  *
*******************************************************************/

BOOL __FsGetNameA(LPCSTR shortname, LPSTR fullname, LPSTR unixname, 
                LPSTR *ppcertain, INT count)
{
   INT    drive;
   LPSTR  proot;
      
   if(!__FsGetFullPathNameA(shortname,fullname, MAX_PATH))
      return FALSE;
   
   drive=isupper(fullname[0])?fullname[0]-'A':fullname[0]-'a';
   
   proot=NULL;
   if(unixname)
   {
      if(__Drive[drive].UnixRoot==NULL)
         return __ErrorReturnFalse(ERROR_INVALID_DRIVE);
         
      proot=unixname+istrcpy(unixname,__Drive[drive].UnixRoot,count)-1;
      istrcat(proot,fullname+2,count);
      __FsChangeSeparator(proot,'/');
   }
   if(ppcertain)
      *ppcertain=proot;
   
   return TRUE;
}

BOOL __FsGetFullPathNameA(LPCSTR pathname, LPSTR fullname, INT buflen)
{  
   INT    subdircount;
   LPSTR  tempstr;   
   LPSTR  fullstr;
   INT    drive;

   /* sanity checks */
   if((pathname==NULL)||(pathname[0]=='\0')||(buflen<4))
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);  
      
   fullname[buflen-1]='\0';
   
   /* make a full pathname, like 'C:\WINDOWS\WIN.INI' */
   if(isalpha(pathname[0])&&(pathname[1]==':'))
   {
      drive=toupper(pathname[0])-'A';
      if(pathname[2]!='\\')
      {
         CHAR drivekey[5]={'=','A',':','\\','\0'};
             
         drivekey[1]=toupper(pathname[0]);
         if(!GetEnvironmentVariableA(drivekey,fullname,buflen))
            istrcpy(fullname,drivekey+1,buflen);
         else
            istrcat(fullname,"\\",buflen);
         istrcat(fullname,pathname+2,buflen);
      }             
      else
      {
         istrcpy(fullname,pathname,buflen);
      }
   }
   else
   {
      drive=toupper(__CurrentDirectoryA[0])-'A';
      if((pathname[0]=='\\')||(pathname[0]=='/'))
      { 
         CHAR drivekey[3]={'A',':','\0'};
         
         drivekey[0]=toupper(__CurrentDirectoryA[0]);
         istrcpy(fullname,drivekey,buflen);
         istrcat(fullname,pathname,buflen);
      }
      else
      {
         istrcpy(fullname,__CurrentDirectoryA,buflen);
         istrcat(fullname,"/",buflen);         
         istrcat(fullname,pathname,buflen);
      }
   }
   /* check buffer for overflow */ 
   if(fullname[buflen-1]!='\0')
      return __ErrorReturnFalse(ERROR_BUFFER_OVERFLOW);
      
   /* now change the separator to what we want */   
   tempstr=fullname;
   while(*tempstr)
   {
      if((*tempstr=='\\')||(*tempstr=='/'))
      {
         *tempstr='\\';
      }
      tempstr++;      
   }
      
   /* now check for '.' and '..' entries, and deal with them */   
   tempstr=fullname+3;
   fullstr=fullname+3;
   subdircount=0;
   while(*tempstr!='\0')
   {
      /* is this a '.' or '..' ??? */
      if(tempstr[0]=='.')
      {
         if((tempstr[1]=='.')&&((tempstr[2]=='\\')||(tempstr[2]=='\0')))
         {
            if(subdircount!=0)
            {
               subdircount--;
               fullstr=fullstr-2;
               while(*fullstr!='\\')
                  fullstr--;
               fullstr++;              
            }
            else
               break;
         }
         else if((tempstr[1]=='\\')||(tempstr[1]=='\0'))
         {
            break;
         }      
      }
      /* no: then copy the stuff */
      do
      {
         *fullstr=*tempstr;
         fullstr++;
         tempstr++;
      }
      while((*tempstr!='\\')&&(*tempstr!='\0'));
      subdircount++;
   }
   return TRUE;
}

DWORD __FsFindNameFlags(DWORD creatmode)
{
  DWORD findname;
  
  findname=FINDNAME_FILE;
  switch(creatmode)
  {
     case CREATE_NEW:
         findname |= FINDNAME_NEW;
         break;
     case CREATE_ALWAYS:
         findname |= FINDNAME_NEW | FINDNAME_EXIST;
         break;
     case OPEN_EXISTING:
         findname |= FINDNAME_EXIST;
         break;
     case OPEN_ALWAYS:
         findname |= FINDNAME_NEW | FINDNAME_EXIST;
         break;
     case TRUNCATE_EXISTING:
         findname |= FINDNAME_EXIST;
         break;
   }
   return findname;
}

int __FsAccessFlags(DWORD faccess)
{
    int retval = 0;

    switch(faccess)
    {
        case GENERIC_READ:
            retval = O_RDONLY;
            break;

        case GENERIC_WRITE:
            retval = O_WRONLY;
            break;

        case (GENERIC_READ | GENERIC_WRITE):
            retval = O_RDWR;
            break;
    }

    return retval;
}

int __FsCreationFlags(DWORD fcreate)
{
    int retval = 0;

    switch(fcreate)
    {
        case CREATE_NEW:
            retval = O_CREAT | O_EXCL;
            break;

        case CREATE_ALWAYS:
            retval = O_CREAT | O_TRUNC;
            break;

        case OPEN_EXISTING:
            retval = 0;
            break;

        case OPEN_ALWAYS:
            retval = O_CREAT;
            break;

        case TRUNCATE_EXISTING:
            retval = O_TRUNC;
            break;
    }
    return retval;
}

DWORD __FsModeToAttrib(int mode)
{
   DWORD attrib;
   
   attrib=0;
   if(mode & S_IFREG)
      attrib= attrib | FILE_ATTRIBUTE_NORMAL;
   if(mode & S_IFDIR)
      attrib= attrib | FILE_ATTRIBUTE_DIRECTORY;
   if((mode & S_IWRITE)==0)
      attrib= attrib | FILE_ATTRIBUTE_READONLY;
            
   return attrib;
}

int __FsAttribToMode(DWORD attrib)
{
   DWORD mode;
   
   mode=0622;
   if(attrib & FILE_ATTRIBUTE_READONLY)
      mode = mode & (~S_IWRITE);
   
   return mode;
}

#ifdef STILL_NEEDED
/* XXXXX:  I doubt this is really right 
 * there are 369 years between 1601 and 1970
 * of those years 3*24+17 have had one day extra
 */

#define  UTC_START  ((365LL*369+3*24+17)*24*60*60)

BOOL __FsUTCToFileTime( ULONG utc, PFILETIME pfiletime )
{
   register ULONGLONG filetime;
     
   filetime= (utc + UTC_START) * 10000;
   pfiletime->dwLowDateTime=LOLONG(filetime);
   pfiletime->dwHighDateTime=HILONG(filetime);
   return TRUE;
}

DWORD __FsFileTimeToUTC( CONST FILETIME *pfiletime)
{
   register ULONGLONG filetime;
   register ULONGLONG utc;
   
   filetime=MAKELONGLONG(pfiletime->dwLowDateTime,
                         pfiletime->dwHighDateTime);
                         
   utc = (filetime/10000) - UTC_START ;
   
   if(HILONG(utc))
      return 0;
   else
      return LOLONG(utc);
}
#endif

/********************************************************************
*                     GetFileAttributesA  --  KERNEL32              *
********************************************************************/

DWORD WINAPI GetFileAttributesA( LPCSTR filename)
{
   struct stat mystat;   
   LPSTR       pcertain;
   CHAR        winname[MAX_PATH];
   CHAR        unixname[MAX_PATH];  
   
   aprintf("GetFileAttributesA( %s )\n", filename );
   
   if(!__FsGetNameA(filename, winname, unixname, &pcertain, MAX_PATH))
      return FALSE;
      
   if(stat(unixname,&mystat)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, MAX_PATH, 
                      FINDNAME_EXIST | FINDNAME_FILE ))
         return 0;
   
      if(stat(unixname,&mystat)==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));      
   }
   return __FsModeToAttrib(mystat.st_mode);
}

/********************************************************************
*                     SetFileAttributesA  --  KERNEL32              *
********************************************************************/

BOOL WINAPI SetFileAttributesA( LPCSTR filename, DWORD attrib)
{
   LPSTR       pcertain;
   CHAR        winname[MAX_PATH];
   CHAR        unixname[MAX_PATH];  
   int         mode;
   
   aprintf("SetFileAttributesA( %s, 0x%lX )\n", filename, attrib);
   
   if(!__FsGetNameA(filename, winname, unixname, &pcertain, MAX_PATH))
      return FALSE;
   
   mode=__FsAttribToMode(attrib);
      
   if(chmod(unixname,mode)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, MAX_PATH,
                      FINDNAME_EXIST | FINDNAME_FILE ))
         return FALSE;
   
      if(chmod(unixname,mode)==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));      
   }
   return TRUE;
}


/********************************************************************
*                    CreateFileA -- KERNEL32                        *
********************************************************************/

HANDLE WINAPI CreateFileA( LPCSTR pname, DWORD access, DWORD share, 
                           LPSECURITY_ATTRIBUTES psa, DWORD create, 
                           DWORD flagsattr, HANDLE htemp )
{
   HANDLE handle;

   aprintf("CreateFileA( %s, 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX, 0x%lX, )\n",
           pname, access, share, (ULONG)psa, create, flagsattr,
           (ULONG) htemp);
      
   if(htemp)
      return __ErrorReturnInvalid(ERROR_INVALID_PARAMETER);

   if(!strcmp(pname,"CONIN$"))
   {

#ifdef NOTYET   
      __ConinOpen(&handle, access, psa);
#else
      return INVALID_HANDLE_VALUE;      
#endif
   }
   else if(!strcmp(pname,"CONOUT$"))
   {
#ifdef NOTYET   
      __ConoutOpen(&handle, access, psa);
#else
      return INVALID_HANDLE_VALUE;   
#endif
   }
   else
   {
      __FileOpen(&handle, pname, access, share, psa, create, flagsattr, htemp);
   }
   return handle;
}

/********************************************************************
*                     CopyFileA  -- KERNEL32                        *
********************************************************************/

BOOL WINAPI CopyFileA(LPCSTR srcname, LPCSTR destname, BOOL failifexist)
{
   INT         nread;
   INT         oflags;   
   INT         srcfd, destfd;
   LPVOID      buffer;
   LPSTR       psrccertain;
   LPSTR       pdestcertain;
   CHAR        srcunixname[MAX_PATH];
   CHAR        srcfullname[MAX_PATH];
   CHAR        destunixname[MAX_PATH];
   CHAR        destfullname[MAX_PATH];
     
   if(!__FsGetNameA(srcname,  srcfullname,  srcunixname,  &psrccertain, 
                  MAX_PATH))
      return FALSE;
   if(!__FsGetNameA(destname, destfullname, destunixname, &pdestcertain, 
                  MAX_PATH))
      return FALSE;
   
   if(failifexist)
      oflags= O_CREAT | O_EXCL | O_WRONLY ;
   else
      oflags= O_CREAT | O_TRUNC | O_WRONLY ;
            
   /* open files and allocate a buffer */
   srcfd=open(srcunixname, O_RDONLY, 0666 );
   if(srcfd==-1) 
   {
      if(!__FsFindNameA(srcfullname, srcunixname, psrccertain, MAX_PATH, 
                      FINDNAME_EXIST |  FINDNAME_FILE ))
         return FALSE;
      
      srcfd=open(srcunixname, O_RDONLY, 0666 );
      if(srcfd==-1) 
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_open_src; }
   }   
   
   destfd=open(destunixname, oflags, 0666 );
   if(destfd==-1)
   {
      if(!__FsFindNameA(destfullname, destunixname, pdestcertain, MAX_PATH, 
                      FINDNAME_FILE ))
         return FALSE;
      
      destfd=open(destunixname, oflags, 0666 );
      if(destfd==-1) 
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_open_dest; }
   }        
      
   buffer=imalloc(__FsBufferSize);
   if(buffer==NULL)
      { SetLastError(ERROR_OUTOFMEMORY); goto cleanup_copy; }
      
   /* copy file */
   do 
   {
      nread=read(srcfd,buffer,__FsBufferSize); 
      if(nread==-1)
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_copy; }
      if(write(destfd,buffer,nread)==-1)
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_copy; }
   }        
   while(nread==__FsBufferSize);
   
   /* close files and free buffer */
   free(buffer);
   close(destfd);
   close(srcfd); 
   
cleanup_copy:
   free(buffer);
cleanup_open_dest:   
   close(destfd);
cleanup_open_src:
   close(srcfd);
   return FALSE;
}

/********************************************************************
*                     MoveFileA  -- KERNEL32                        *
********************************************************************/

BOOL WINAPI MoveFileA(LPCSTR srcname, LPCSTR destname)
{
   INT         nread;
   INT         srcfd, destfd;
   LPVOID      buffer;
   LPSTR       psrccertain;
   LPSTR       pdestcertain;
   CHAR        srcunixname[MAX_PATH];
   CHAR        srcfullname[MAX_PATH];
   CHAR        destunixname[MAX_PATH];
   CHAR        destfullname[MAX_PATH];
     
   if(!__FsGetNameA(srcname,  srcfullname,  srcunixname,  &psrccertain, 
                  MAX_PATH))
      return FALSE;
   if(!__FsGetNameA(destname, destfullname, destunixname, &pdestcertain, 
                  MAX_PATH))
      return FALSE;
   
   /* first try rename */
   if(!rename(srcunixname, destunixname))
      return TRUE;
   
   if(!__FsFindNameA(destfullname, destunixname, pdestcertain, MAX_PATH, 
                   FINDNAME_EXIST | FINDNAME_FILE ))
       return FALSE;
         
   if(!__FsFindNameA(destfullname, destunixname, pdestcertain, MAX_PATH, 
                   FINDNAME_NEW | FINDNAME_FILE ))
       return FALSE;
         
   if(!rename(srcunixname, destunixname))
      return TRUE;
   
   if(errno!=EXDEV)
      return __ErrorReturnFalse(__ErrorFromErrno(errno));
   
   /* open files and allocate a buffer */
   srcfd=open(srcunixname, O_RDONLY, 0666 );
   if(srcfd==-1) 
      { SetLastError(__ErrorFromErrno(errno)); goto cleanup_open_src; }
  
   
   destfd=open(destunixname, O_CREAT | O_EXCL | O_WRONLY,  0666 );
      { SetLastError(__ErrorFromErrno(errno)); goto cleanup_open_dest; }
      
   buffer=imalloc(__FsBufferSize);
   if(buffer==NULL)
      { SetLastError(ERROR_OUTOFMEMORY); goto cleanup_copy; }
      
   /* copy file */
   do 
   {
      nread=read(srcfd,buffer,__FsBufferSize); 
      if(nread==-1)
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_copy; }
      if(write(destfd,buffer,nread)==-1)
         { SetLastError(__ErrorFromErrno(errno)); goto cleanup_copy; }
   }        
   while(nread==__FsBufferSize);
      
   /* close files and free buffer */
   free(buffer);
   close(destfd);
   close(srcfd); 
   
   /* remove srcfile */
   if(remove(srcunixname))
      return TRUE;
   else
   {      
      remove(destunixname);       
      return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }
   
cleanup_copy:
   free(buffer);
cleanup_open_dest:   
   close(destfd);
cleanup_open_src:
   close(srcfd);
   return FALSE;
}

/********************************************************************
*                     DeleteFileA  -- KERNEL32                      *
********************************************************************/
 
BOOL WINAPI DeleteFileA(LPCSTR filename)
{
   LPSTR pcertain;
   CHAR  winname[MAX_PATH];
   CHAR  unixname[MAX_PATH];
        
   if(!__FsGetNameA(filename, winname, unixname, &pcertain, MAX_PATH));               
      return FALSE;
   
   /* delete file */
   if(remove(unixname)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, 
                      MAX_PATH, FINDNAME_FILE | FINDNAME_EXIST ))
         return FALSE;

      if(remove(unixname)==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }   
   return TRUE; 
}


/*********************************************************************
*                  GetFullPathNameA  --  KERNEL32                    *
*********************************************************************/

DWORD WINAPI GetFullPathNameA(LPCSTR filename, DWORD buflen, LPSTR buffer,
                              LPSTR *pfilepart)
{   
   INT retval;
   
   retval=__FsGetFullPathNameA(filename, buffer, buflen);
   
   if(pfilepart)
   {
      *pfilepart=strrchr(buffer, '\\');
   }
   return retval;   
}                             

/********************************************************************
*                  SetCurrentDirectoryA  --  KERNEL32               *
********************************************************************/

BOOL WINAPI SetCurrentDirectoryA(LPCSTR pathname)
{
   LPSTR pcertain;
   CHAR  unixname[MAX_PATH];
   CHAR  winname[MAX_PATH];
   
   /* get a unixname */
   if(!__FsGetNameA(pathname, winname, unixname, &pcertain, MAX_PATH))
      return FALSE;   
   
   /* try chdir */
   if(chdir(unixname)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, MAX_PATH, 
                      FINDNAME_EXIST | FINDNAME_DIR ))
         return FALSE;
         
      if(chdir(unixname)==-1)      
         return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }   
   /* set unix directory   */
   strcpy(__UnixCurrentDirectory,unixname);
   
   /* set win32 directory  */
   strcpy(__CurrentDirectoryA,winname);
   
   /* set current drive and the drive's current path */   
   
   if(strlen(winname)>3)
   {
      CHAR drivekey[5]={'=','A',':','\\','\0'};
      drivekey[1]=toupper(winname[0]);
      
      SetEnvironmentVariableA(drivekey,winname);
   }
   return TRUE;   
}

/********************************************************************
*                  GetCurrentDirectoryA  --  KERNEL32               *
********************************************************************/

DWORD WINAPI GetCurrentDirectoryA(DWORD buflen, LPSTR buffer)
{
   return istrcpy(buffer, __CurrentDirectoryA, buflen);
}

/********************************************************************
*                  CreateDirectoryA  -- KERNEL32                    *
********************************************************************/
 
BOOL WINAPI CreateDirectoryA(LPCSTR dirname, LPSECURITY_ATTRIBUTES psa)
{
   LPSTR pcertain;
   CHAR  winname[MAX_PATH];
   CHAR  unixname[MAX_PATH];
        
   if(!__FsGetNameA(dirname, winname, unixname, &pcertain, MAX_PATH));               
      return FALSE;
   
   /* delete file */
   if(mkdir(unixname, 0666)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, 
                      MAX_PATH, FINDNAME_DIR | FINDNAME_NEW ))
         return FALSE;

      if(mkdir(unixname, 0666)==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }   
   return TRUE; 
}
/********************************************************************
*                     RemoveDirectoryA  -- KERNEL32                 *
********************************************************************/
 
BOOL WINAPI RemoveDirectoryA(LPCSTR dirname)
{
   LPSTR pcertain;
   CHAR  winname[MAX_PATH];
   CHAR  unixname[MAX_PATH];
        
   if(!__FsGetNameA(dirname, winname, unixname, &pcertain, MAX_PATH));               
      return FALSE;
   
   /* delete file */
   if(rmdir(unixname)==-1)
   {
      if(!__FsFindNameA(winname, unixname, pcertain, 
                      MAX_PATH, FINDNAME_FILE | FINDNAME_EXIST ))
         return FALSE;

      if(rmdir(unixname)==-1)
         return __ErrorReturnFalse(__ErrorFromErrno(errno));
   }   
   return TRUE; 
}

/********************************************************************
*                     GetVolumeInformationA  -- KERNEL32            *
********************************************************************/

BOOL WINAPI GetVolumeInformationA(LPCSTR root, LPSTR volname,
                                  DWORD volnamesize, LPDWORD psernum,
                                  LPDWORD pmaxpath, LPDWORD pfsflags,
                                  LPSTR fsname, DWORD fsnamesize)
{
   aprintf("GetVolumeInformationA"
           "( %s, 0x%lX, %ld, 0x%lX, 0x%lX, 0x%lX, 0x%lX, %ld )\n",
            root, (ULONG) volname, volnamesize, (ULONG) psernum,
            (ULONG) pmaxpath, (ULONG) pfsflags, (ULONG) fsname, fsnamesize );
            
   /* XXX: hacko hack */
   istrcpy(volname,"LINUX",volnamesize);
   istrcpy(fsname,"LINUX",fsnamesize);
   *psernum=0;
   *pmaxpath=MAX_PATH;
   *pfsflags=FS_CASE_IS_PRESERVED;
   return TRUE; 
}                                  