/*u
 * kernel/heap.c
 * Copyright (C) 1996, Onno Hovers, All rights reserved
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; see the file COPYING.LIB. If
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 * Cambridge, MA 02139, USA.  
 *
 * Win32 heap functions (HeapXXX, GlobalXXX, LocalXXX). 
 * 
 */

/* 
 * Win32 allows different heaps to be allocated. So for those programs
 * that use Win32 heap functions (HeapXXX, GlobalXXX, LocalXXX), I need 
 * a good heap algorithm.  
 *
 * alloc. size	admin size:	size of subblock (+hdr)  allocations 
 * 16		32		1K			 31             
 * 32		48		1K			  
 * 48           64 		1K
 * 64	        80		  
 * 128		144  
 * 256		272		16 
 */

#include <stdlib.h>
#include <win32/defines.h>
#include <win32/kernel.h>
#include <win32/error.h>
#include "virtual.h"
#include "wfc.h"
#include "thread.h"
#include "object.h"
#include "error.h"

/* definitions */
#define HEAP_ADMIN_SIZE   8
#define HEAP_ROUNDVAL     (2*(HEAP_ADMIN_SIZE)-1) 
#define HEAP_FRAGMENT_THRESHOLD 256

#define HEAP_FRAG_MAGIC   0xF0
#define HEAP_ALLOC_MASK   0xF0000000
#define HEAP_FREE_MASK    0x80000000
#define HEAP_SIZE_MASK    0x0FFFFFFF
#define HEAP_FREE_TAG     0x80000000
#define HEAP_FRAG_TAG     0x10000000
#define HEAP_NORMAL_TAG   0x30000000
#define HEAP_OLD_TAG      0x50000000
#define HEAP_SUB_TAG      0x70000000

#define HEAP_ISFREE(p)    ((((PHEAP_BLOCK)p)->Size) & HEAP_FREE_MASK)
#define HEAP_ISALLOC(p)   (((((PHEAP_BLOCK)p)->Size) & HEAP_FREE_MASK)==0)
#define HEAP_ISFRAG(p)    (((((PHEAP_BLOCK)p)->Size) & HEAP_ALLOC_MASK)\
                          ==HEAP_FRAG_TAG)
#define HEAP_ISNORMAL(p)  (((((PHEAP_BLOCK)p)->Size) & HEAP_ALLOC_MASK)\
                          ==HEAP_NORMAL_TAG)
#define HEAP_ISSUB(p)     (((((PHEAP_BLOCK)p)->Size) & HEAP_ALLOC_MASK)\
                          ==HEAP_SUB_TAG)
#define HEAP_ISOLD(p)     (((((PHEAP_BLOCK)p)->Size) & HEAP_ALLOC_MASK)\
                          ==HEAP_OLD_TAG)

#define HEAP_SIZE(p)      ((((PHEAP_BLOCK)p)->Size) & HEAP_SIZE_MASK )
#define HEAP_RSIZE(p)     ((((PHEAP_BLOCK)p)->Size  + HEAP_ROUNDVAL)\
                          & HEAP_SIZE_MASK )
#define HEAP_PREVSIZE(p)  ((((PHEAP_BLOCK)p)->PrevSize) & HEAP_SIZE_MASK )
                          
#define HEAP_PREV(p)      ((PHEAP_BLOCK)((LPVOID)p+HEAP_PREVSIZE(p)))
#define HEAP_NEXT(p)      ((PHEAP_BLOCK)((LPVOID)p+HEAP_RSIZE(p)))

typedef struct __HEAP_BLOCK
{
   ULONG  Size;            /* this is relative to Data */ 
   ULONG  PrevSize;        /* p - p->PrevSize is the previous block */ 
   union
   {
      BYTE  Data[ANYSIZE_ARRAY];
      struct 
      {
        HANDLE Handle;
        BYTE  Data[ANYSIZE_ARRAY];
      } Old;
   } Use;
} HEAP_BLOCK, *PHEAP_BLOCK;

struct __HEAP_SUBALLOC;

typedef struct __HEAP_FRAGMENT
{  
   UBYTE			 Magic;
   UBYTE			 Number;
   UWORD			 Size;
   struct __HEAP_SUBALLOC 	*Sub;
   
   /* this is only used in free blocks */
   struct __HEAP_FRAGMENT	*FreeNext;
   struct __HEAP_FRAGMENT	*FreePrev;
} HEAP_FRAGMENT, *PHEAP_FRAGMENT, *LPHEAP_FRAGMENT;

typedef struct __HEAP_SUBALLOC
{
   UWORD			 Magic;
   UWORD			 NumberFree;   
   struct __HEAP_SUBALLOC	*Next;   

   struct __HEAP_SUBALLOC	*Prev;   
   struct __HEAP_FRAGMENT	*FirstFree;
   struct __HEAP_BUCKET		*Bucket;       
   ULONG			 Bitmap; 
} HEAP_SUBALLOC, *PHEAP_SUBALLOC, *LPHEAP_SUBALLOC;

typedef struct __HEAP_BUCKET
{
   struct __HEAP_SUBALLOC	*FirstFree;
   UWORD			 Size;
   UWORD			 Number;
   UWORD			 TotalSize;
} HEAP_BUCKET, *PHEAP_BUCKET, *LPHEAP_BUCKET;

typedef struct __HEAP
{
   ULONG 		Magic;
   LPVOID 		End;  

   ULONG		Flags;
   CRITICAL_SECTION	Synchronize;

   HEAP_BUCKET		Bucket[8];
      
   struct __HEAP	*NextHeap;
   /* this has to aligned on an 8 byte boundary */
   HEAP_BLOCK		Start  __attribute__((aligned (8)));
} HEAP, *PHEAP;

PHEAP	__ProcessHeap;

static HEAP_BUCKET __HeapDefaultBuckets[]=
{      
  { NULL,  16, 18, 504  },
  { NULL,  24, 30, 1016 },
  { NULL,  32, 24, 1016 },
  { NULL,  48, 17, 1016 },    
  { NULL,  64, 27, 2040 },
  { NULL,  96, 19, 2040 },
  { NULL, 128, 29, 4088 },
  { NULL, 256, 15, 4088 }, 
};

/*********************************************************************
*                      __HeapAlloc                                   *
*                                                                    *
* allocates a range of memory from the heap                          *
*********************************************************************/

static LPVOID __HeapAlloc(PHEAP pheap, ULONG size, ULONG tag)
{
   PHEAP_BLOCK	pfree;
   PHEAP_BLOCK	palloc;
   PHEAP_BLOCK	pnext;
   LPVOID	commitstart;
   LPVOID	commitend;
   ULONG	freesize;
   ULONG	allocsize;
   
   pfree=&(pheap->Start);
   allocsize=ROUNDUP(size,8);
   freesize=HEAP_SIZE(pfree);
   /* look for a free region of memory: simple First Fit */
   while( HEAP_ISALLOC(pfree) || ( freesize<allocsize ))
   {
      pfree=HEAP_NEXT(pfree);
      if((LPVOID) pfree>=pheap->End)
         return NULL;
      freesize=HEAP_SIZE(pfree);
   }
   palloc=pfree;
   freesize=HEAP_SIZE(pfree);
     
   if(freesize>allocsize)
   {
      /* commit necessary memory */
      commitstart=(LPVOID) ROUNDUP((ULONG) palloc+HEAP_ADMIN_SIZE,PAGE_SIZE);
      commitend  =(LPVOID) ROUNDDOWN((ULONG) palloc+
                                     allocsize+2*HEAP_ADMIN_SIZE, PAGE_SIZE);
      if(commitstart<commitend)
         __VirtualCommit(commitstart, commitend-commitend, PAGE_READWRITE);

      /* split this block in two */
      pfree= (LPVOID) pfree+ HEAP_ADMIN_SIZE + allocsize;

      /* update admin */
      pfree->Size     = (palloc->Size-allocsize-HEAP_ADMIN_SIZE) & HEAP_FREE_TAG ; 
      pfree->PrevSize = pfree-palloc;
            
      pnext=HEAP_NEXT(pfree);
      if((LPVOID) pnext < pheap->End );
         pnext->PrevSize=pfree->Size+HEAP_ADMIN_SIZE;
   }
   else
   {  
      /* commit necessary memory */
      commitstart=(LPVOID) ROUNDUP((ULONG) palloc+HEAP_ADMIN_SIZE, PAGE_SIZE);
      commitend  =(LPVOID) ROUNDDOWN((ULONG) palloc+HEAP_ADMIN_SIZE +allocsize,
                           PAGE_SIZE);
      if(commitstart<commitend)
         __VirtualCommit(commitstart, commitend-commitend, PAGE_READWRITE);
   }
   /* update our administration */   
   palloc->Size= size & tag;
   return palloc;
}

static BOOL __HeapFree(PHEAP pheap, LPVOID ptr)
{
   PHEAP_BLOCK	pfree=((PHEAP_BLOCK)ptr)-1;
   PHEAP_BLOCK	pprev;
   PHEAP_BLOCK	pnext;
   LPVOID	decommitstart;
   LPVOID	decommitend;
   
   /* check that this is a valid allocated block */
   if(HEAP_ISALLOC(pfree))
      return FALSE;
   
   pfree->Size = ROUNDUP(pfree->Size,8) & HEAP_FREE_TAG;

   /* if there is a free region of memory before us, join it */
   pprev=HEAP_PREV(pfree);
   pnext=HEAP_NEXT(pfree);
   if((pprev!=pfree) && HEAP_ISFREE(pprev))
   {
      pprev->Size +=pfree->Size + HEAP_ADMIN_SIZE;
       if((LPVOID) pnext<pheap->End)
         pnext->PrevSize=pnext-pprev;

      pfree->Size=0;
      pfree->PrevSize=0;      
      pfree=pprev;
   }   
   /* if there is a free region of memory after us, join it */
   if(((LPVOID) pnext< pheap->End)&& HEAP_ISFREE(pnext))
   {
      pfree->Size += pnext->Size + HEAP_ADMIN_SIZE;

      pnext->Size=0;
      pnext->PrevSize=0;
      pnext=HEAP_NEXT(pfree);

      if((LPVOID) pnext< pheap->End)
         pnext->PrevSize=pnext-pfree;
   }
   
   /* decommit unnecessary memory */
   decommitstart=(LPVOID) ROUNDUP((ULONG) pfree+HEAP_ADMIN_SIZE ,PAGE_SIZE);
   decommitend  =(LPVOID) ROUNDDOWN((ULONG) pfree+HEAP_ADMIN_SIZE+
                                    HEAP_SIZE(pfree), PAGE_SIZE);
   if(decommitstart<decommitend)
      __VirtualDecommit(decommitstart, decommitend-decommitend);

   return TRUE;
}

PHEAP_SUBALLOC __HeapAllocSub(PHEAP pheap, PHEAP_BUCKET pbucket)
{
   INT			i;
   INT			add;
   PHEAP_SUBALLOC	psub;
   PHEAP_FRAGMENT	pprev;
   PHEAP_FRAGMENT	pnext;
   PHEAP_FRAGMENT	palloc;
       
   psub=(PHEAP_SUBALLOC) __HeapAlloc(pheap, pbucket->TotalSize,
                                     HEAP_SUB_TAG);                                        
   if(!psub)
      return __ErrorReturnNull(ERROR_OUTOFMEMORY);           
      
   /* initialize suballoc */
   palloc=(PHEAP_FRAGMENT) ((LPVOID)psub + sizeof(PHEAP_SUBALLOC)); 
   psub->FirstFree=palloc;
   psub->NumberFree=pbucket->Number;
   psub->Bitmap=0;     
   psub->Next=pbucket->FirstFree;
   psub->Prev=NULL;
   pbucket->FirstFree=psub;
      
   /* initialize free fragments */
   add=pbucket->Size+sizeof(HEAP_FRAGMENT);
   pprev=NULL;
   for(i=0;i<pbucket->Number;i++)
   {
      pnext=(PHEAP_FRAGMENT)((LPVOID)palloc+add);
      palloc->Magic=HEAP_FRAG_MAGIC;
      palloc->Number=0;
      palloc->Size=pbucket->Size; 
      palloc->Sub=psub;  
      palloc->FreeNext=pnext;
      palloc->FreePrev=pprev;          
      pprev=palloc;
      palloc=pnext;
      i++;
   }
   pprev->FreeNext=NULL;
   return psub;
}

static LPVOID __HeapAllocFragment(PHEAP pheap, ULONG size )
{
   PHEAP_BUCKET		pbucket;
   PHEAP_SUBALLOC	psub;
   PHEAP_FRAGMENT	palloc;
   INT			nalloc;
   
   /* get bucket size */
   pbucket=pheap->Bucket;  
   while(size<pbucket->Size)
   {   
      pbucket++;
   }  
   /* get suballoc */   
   psub   = pbucket->FirstFree;
   if(!psub)
      psub = __HeapAllocSub(pheap, pbucket);
   if(!psub)
      return NULL;
     
   /* do our bookkeeping */    
   palloc = psub->FirstFree;
   psub->FirstFree = palloc->FreeNext;
   nalloc = palloc->Number;
   psub->NumberFree--;
   psub->Bitmap|=(1<<nalloc);
   
   /* advance freelist */
   if(!psub->NumberFree)
      pbucket->FirstFree=psub->Next;
   
   /* initialize allocated block */
   palloc->Magic=HEAP_FRAG_MAGIC;
   palloc->Size=size;
   return palloc+8;
}

static BOOL __HeapFreeFragment(PHEAP pheap, LPVOID pfree )
{
   PHEAP_BUCKET		pbucket;
   PHEAP_SUBALLOC	psub;
   PHEAP_FRAGMENT	pfrag=((PHEAP_FRAGMENT)pfree)-1;
   INT			nalloc;
   
   /* sanity checks */
   if(pfrag->Magic!=HEAP_FRAG_MAGIC)
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);
      
   /* get bucket size */
   psub=pfrag->Sub;
   pbucket=psub->Bucket;
   
   nalloc=pfrag->Number;
   if((psub->Bitmap&(1<<nalloc))==0)
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);
   psub->NumberFree++;
   if(psub->NumberFree==0)
   {
      /* free suballoc */
      if(psub==pbucket->FirstFree)
         pbucket->FirstFree=psub->Next;
      if(psub->Prev)
         psub->Prev->Next=psub->Next;
      if(psub->Next)
         psub->Next->Prev=psub->Prev;
      if(!__HeapFree(pheap, psub))
         return FALSE;
   }
   else
   {
      /* free fragment */
      psub->Bitmap&= ~(1<<nalloc);
      if(pfrag==psub->FirstFree)
         psub->FirstFree=pfrag->FreeNext;
      if(pfrag->FreePrev)
         pfrag->FreePrev->FreeNext=pfrag->FreeNext;
      if(pfrag->FreeNext)
         pfrag->FreeNext->FreePrev=pfrag->FreePrev;     
   }
   return TRUE;   
}

/*********************************************************************
*                     __HeapPrepare                                  *
*                                                                    *
* Fills in all the data structures of a heap                         * 
*********************************************************************/

PHEAP WINAPI __HeapPrepare(LPVOID base, ULONG minsize, ULONG maxsize, 
                           ULONG flags)
{
   PHEAP pheap=(PHEAP) base; 
       
   pheap->Magic=MAGIC_HEAP;
   pheap->End= ((LPVOID)pheap)+minsize;
   pheap->Flags=flags;
   memcpy(pheap->Bucket,__HeapDefaultBuckets,sizeof(__HeapDefaultBuckets));
   if(__ProcessHeap)
   {
      pheap->NextHeap=__ProcessHeap->NextHeap;
      __ProcessHeap->NextHeap=pheap; 
   }      
   else
   {
      pheap->NextHeap=0;
      __ProcessHeap=pheap;     
   } 
   InitializeCriticalSection(&(pheap->Synchronize));
   pheap->Start.Size= minsize-sizeof(HEAP)-HEAP_ADMIN_SIZE;
   pheap->Start.PrevSize =0;
   
   return pheap;
}

/*********************************************************************
*                     __HeapInit                                     *
*                                                                    *
* Called by __VirtualInit to initialize the default process heap     *
*********************************************************************/

VOID WINAPI __HeapInit(LPVOID base, ULONG minsize, ULONG maxsize)
{
   mmap(base, PAGE_SIZE, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, 
        __DevZero, 0);
   
   __HeapPrepare(base, minsize, maxsize, 0);  
}


/*********************************************************************
*                     HeapCreate -- KERNEL32                         *
*********************************************************************/

HANDLE WINAPI HeapCreate(ULONG flags, ULONG minsize, ULONG maxsize)
{
   PHEAP pheap;
   
   pheap = __VirtualReserve(NULL, minsize, PAGE_READWRITE);
   __VirtualCommit(pheap, PAGE_SIZE, PAGE_READWRITE);
      
   return (HANDLE) __HeapPrepare(pheap, minsize, maxsize, flags);
}

/*********************************************************************
*                     HeapDestroy -- KERNEL32                        *
*********************************************************************/

BOOL WINAPI HeapDestroy(HANDLE hheap)
{
   PHEAP pheap=(PHEAP) hheap;
   
   if(pheap->Magic!=MAGIC_HEAP)
      return __ErrorReturnFalse(ERROR_INVALID_PARAMETER);
      
   DeleteCriticalSection(&(pheap->Synchronize));
   __VirtualRelease(pheap);   

   return TRUE;
}

/*********************************************************************
*                     HeapAlloc -- KERNEL32                          *
*********************************************************************/
 
LPVOID WINAPI HeapAlloc(HANDLE hheap, ULONG size, ULONG flags)
{   
   PHEAP    pheap=hheap;
   LPVOID   retval;
   
   if(( flags | pheap->Flags)  & HEAP_NO_SERIALIZE )
      EnterCriticalSection(&(pheap->Synchronize));

   if(size>HEAP_FRAGMENT_THRESHOLD)
      retval=__HeapAlloc(pheap, size, HEAP_NORMAL_TAG);   
   else
      retval=__HeapAllocFragment(pheap, size);      

   if( (flags | pheap->Flags) & HEAP_NO_SERIALIZE )
      LeaveCriticalSection(&(pheap->Synchronize));

   return retval;

}


/*********************************************************************
*                     HeapFree -- KERNEL32                           *
*********************************************************************/

BOOL WINAPI HeapFree(HANDLE hheap, ULONG flags, LPVOID ptr)
{
   PHEAP            pheap=hheap;
   PHEAP_BLOCK      pfree=((PHEAP_BLOCK)ptr-1);   
   BOOL             retval;
   
   if(( flags | pheap->Flags)  & HEAP_NO_SERIALIZE )
      EnterCriticalSection(&(pheap->Synchronize));

   if(HEAP_ISNORMAL(pfree))
      retval=__HeapFree(pheap, ptr);
   else if(HEAP_ISFRAG(pfree))
      retval=__HeapFreeFragment(pheap, ptr);
   else
      retval=__ErrorReturnFalse(ERROR_INVALID_PARAMETER);

   if( (flags| pheap->Flags) & HEAP_NO_SERIALIZE )
      LeaveCriticalSection(&(pheap->Synchronize));

   return retval;
}

/*********************************************************************
*                   GetProcessHeap  --  KERNEL32                     *
*********************************************************************/

HANDLE WINAPI GetProcessHeap(VOID)
{
   return (HANDLE) __ProcessHeap;
}

/********************************************************************
*                   GetProcessHeaps  --  KERNEL32                   *
*                                                                   *
* NOTE in Win95 this function is not implemented and just returns   *
* ERROR_CALL_NOT_IMPLEMENTED                                        * 
********************************************************************/

DWORD WINAPI GetProcessHeaps(DWORD maxheaps, PHANDLE phandles )
{
   DWORD retval;
   PHEAP pheap;
   
   THREAD_ENTER
   
      pheap=__ProcessHeap;
      retval=0;   
      while((pheap)&&(maxheaps))
      {
         *phandles=pheap;
         phandles++;
         maxheaps--;
         retval++;     
         pheap=pheap->NextHeap;     
      }
      while(pheap)
      {
         retval++;
         pheap=pheap->NextHeap;
      }
   
   THREAD_LEAVE  
    
   return retval;
}

/*********************************************************************
*                    HeapLock  --  KERNEL32                          *
*********************************************************************/

BOOL WINAPI HeapLock(HANDLE hheap)
{
   PHEAP pheap=hheap;

   EnterCriticalSection(&(pheap->Synchronize));
   return TRUE;
}

/*********************************************************************
*                    HeapUnlock  --  KERNEL32                        *
*********************************************************************/

BOOL WINAPI HeapUnlock(HANDLE hheap)
{
   PHEAP pheap=hheap;

   LeaveCriticalSection(&(pheap->Synchronize));
   return TRUE;
}

#ifdef NOTYET
/*********************************************************************
*                    HeapCompact  --  KERNEL32                       *
*********************************************************************/

UINT HeapCompact(HANDLE hheap, DWORD flags)
{
   PHEAP  pheap=hheap;
   UINT   retval;
   
   /* we keep the heap as compact as we can, so this is a nop */
#ifdef NEVER   
   if(( flags | pheap->Flags)  & HEAP_NO_SERIALIZE )
      EnterCriticalSection(&(pheap->Synchronize));
#endif

   retval = pheap->LargestCommit;

#ifdef NEVER
   if( (flags| pheap->Flags) & HEAP_NO_SERIALIZE )
      LeaveCriticalSection(&(hheap->Synchronize));
#endif   
   
   return retval;   
}

/*********************************************************************
*                    HeapSize  --  KERNEL32                          *
*********************************************************************/

DWORD WINAPI HeapSize(HANDLE hheap, DWORD flags, LPVOID pmem)
{
   PHEAP   pheap=(PHEAP) hheap;
   PALLOC  palloc;
   
   if(pheap->Magic!=MAGIC_HEAP)
      { SetLastError(ERROR_INVALID_PARAMETER); return 0; }
      
   if(( flags | pheap->Flags)  & HEAP_NO_SERIALIZE )
      EnterCriticalSection(&(pheap->Synchronize));

   if((ptr>pheap)||(ptr<pheap->End))
   {
      palloc=((PHEAP_BLOCK) pmem)-1;
     
      if(!HEAP_ISALLOC(palloc))
         { SetLastError(ERROR_INVALID_PARAMETER); retval = 0; }
      else
         retval=HEAP_SIZE(palloc->Size);
   }
 
   if( (flags| pheap->Flags) & HEAP_NO_SERIALIZE )
       LeaveCriticalSection(&(hheap->Synchronize));
   
   return retval;   
}

/*********************************************************************
*                    GlobalAlloc  --  KERNEL32                       *
*********************************************************************/

HGLOBAL GlobalAlloc(UINT flags, DWORD size)
{ 
    LPVOID palloc;
    
    if(flags & GMEM_FIXED)
    {
        palloc=HeapAlloc(__Heap, size, 0);            
        return (HGLOBAL) palloc;
    }
    else
    {
        palloc=HeapAlloc(__Heap, size, 0);
        phandle=__HeapAllocHandle(__Heap);               
        phandle->Pointer=palloc;
        phandle->Magic=MAGIC_HEAP_MOVABLE_HANDLE;
        phandle->Flags=(flags&GMEM_DISCARDABLE)?2:0;
        phandle->LockCount=0;
        return (HGLOBAL) phandle; 
    }
}

/*********************************************************************
*                    GlobalFree  --  KERNEL32                        *
*********************************************************************/

HGLOBAL GlobalFree(HGLOBAL hmem)
{    
    if((hmem&4)==0)
    {
        if(!HeapFree(__Heap, (LPVOID) hmem, 0)
           return hmem;
        else
           return NULL;         
    }
    else
    {
        if(phnd->Magic==MAGIC_HEAP_MOVABLE_HANDLE)
            phnd->                                                    
    }    
} 
#endif