/*
 * thread support functions
 * Copyright (C) 1996, Onno Hovers, All rights reserved
 *
 * we enter a global kind of critical section for
 * synchronization, if the system supports kernel threads
 *  
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; see the file COPYING.LIB. If
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 * Cambridge, MA 02139, USA.  
 */

#include <win32/defines.h>
#include <win32/kernel.h>
#include <i386/segment.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <linux/unistd.h>
#include "teb.h"
#include "imalloc.h"

#ifdef NOTYET
/* lock a thread */
#define THREAD_ENTER   { if( __MultiThreading ) __ThreadEnter();   }
#define THREAD_LEAVE   { if( __MultiThreading ) __ThreadLeave();   }

/* temporary unlock a thread before entering a blocking operation */
#define THREAD_UNBLOCK(p) { if( __MultiThreading ) __ThreadUnblock(p); } 
#define THREAD_BLOCK(p)   { if( __MultiThreading ) __ThreadBlock(p);   }
 
/* actually the number of threads - 1, zero if only one thread runs */
INT __MultiThreading=0;
INT __ThreadLockCount=-1; 	
INT __ThreadRecurse=0;
INT __ThreadLockId=-1;
INT __ThreadSemaphore=-1;

#endif

#ifdef __i386__
/* initialize per thread data */
BOOL __ThreadInitTeb(VOID)
{
   PTEB pteb;
   INT  fs;
   
   pteb=imalloc(sizeof(TEB));
   pteb->Except   =(PVOID) -1;
   pteb->TlsArray =NULL; 
   
   fs=__SegmentAlloc(pteb,sizeof(TEB),
                     SEGMENT_DATA | SEGMENT_WRITABLE | SEGMENT_32BIT );
   
   if(!fs)
      return FALSE;
                  
   __asm__ __volatile__ ( "\t movw  %w0,%%fs \n" :: "r" (fs) );
   
   return TRUE;
} 
#endif

#ifdef NOTYET

VOID __ThreadInit()
{
   __ThreadSemaphore=semget(IPC_PRIVATE,1,IPC_CREAT);   
}

VOID __ThreadDone()
{
   semctl(__ThreadSemaphore,0,IPC_RMID,NULL)
}

VOID __ThreadEnter(VOID)
{
   if( __MultiThreading )
   {     
      if( InterlockedIncrement(&__ThreadLockCount) )
      {
         if( __ThreadLockId!=mt_getpid() )
         {                 
            struct sembuf sop;
                                                
            sop.sem_num=0;
            sop.sem_op=0;
            sop.sem_flg=0;            
            semop(__ThreadSemaphore,&sop,0);
            
            __ThreadLockId=mt_getpid();                
         }
      }
      else
      {      
         __ThreadLockId=mt_getpid();
      }
      __ThreadLockRecurse++;            
   }
}

VOID __ThreadLeave(VOID)
{
  if( __MultiThreading )
  { 
    if( __ThreadLockId!=mt_getpid())
       return;
       
    __ThreadLockRecurse--;
    if(__ThreadLockRecurse==0)
    {
       __ThreadLockId=-1;
       if(InterlockedDecrement(&__ThreadLockCount))
       {         
          struct sembuf sop;
          
          sop.sem_num=0;
          sop.sem_op=1;
          sop.sem_flg=0;
          semop(__ThreadSemaphore,&sop,0)       
       }
    }
    else
    {
       InterlockedDecrement(__ThreadLockCount);
    }
  }
}


VOID __ThreadUnblock(PULONG prec)
{
   if( __MultiThreading )
   {
      *prec=0; 
      if( __ThreadLockId!=mt_getpid())
         return;
 
      *prec=__ThreadLockRecurse;      
      __ThreadLockRecurse=0;

      __ThreadLockId=-1;
      for(i=0;i<*prec;i++)
      {    
         if(InterlockedDecrement(&__ThreadLockCount))
         {         
            struct sembuf sop;
          
            sop.sem_num=0;
            sop.sem_op=1;
            sop.sem_flg=0;
            semop(__ThreadSemaphore,&sop,0)
            return;       
         }         
      }         
   }
}

VOID __ThreadBlock(PULONG prec)
{
   
   if( __MultiThreading )
   {  
      INT rec;
   
      rec=*prec;
      /* first get the lock */
      if( InterlockedIncrement(&__ThreadLockCount) )
      {
         struct sembuf sop;
                                                
         sop.sem_num=0;
         sop.sem_op=0;
         sop.sem_flg=0;            
         semop(__ThreadSemaphore,&sop,0);
         
         rec--;
      }      
      __ThreadLockId=mt_getpid();
      __ThreadLockRecurse=*prec;            
     
      while(rec>0)      
      {
         rec--;
         InterlockedIncrement(&__ThreadLockCount);
      }
   }
}


#define CSIGNAL         0x000000ff      /* signal mask to be sent at exit */
#define CLONE_VM        0x00000100      /* set if VM shared between processes */
#define CLONE_FS        0x00000200      /* set if fs info shared between processes */
#define CLONE_FILES     0x00000400      /* set if open files shared between processes */
#define CLONE_SIGHAND   0x00000800      /* set if signal handlers shared */

typedef struct __THREAD_NEW
{
   LPSECURITY_ATTRIBUTES	Security;
   LPVOID			StackHigh;
   LPVOID			StackLow;
   PTHREAD_START_ROUTINE 	Function;
   LPVOID			Paramater;
   ULONG			Flags;  
} THREAD_NEW, *PTHREAD_NEW;

HANDLE __ThreadCreate(LPSECURITY_ATTRIBUTES psa, ULONG stacksize, 
                    LPTHREAD_START_ROUTINE pstart, LPVOID pparam, 
                    ULONG creatflags, PULONG pthreadid)
{
   long retval;
   void **newstack;
   
   THREAD_ENTER; 

   /* allocate new stack for subthread */
   newstack = __VirtualAllocStack(stacksize);
   if(!newstack)
      return NULL;

   /* Set up the stack for child function */
   stackpointer          = newstack+stacksize-sizeof(THREAD_NEW);
   pthreadnew            = (PTHREAD_NEW) stackpointer;
   pthreadnew->Security  = psa;
   pthreadnew->StackHigh = newstack+stackhigh;
   pthreadnew->StackLow  = newstack;
   pthreadnew->Function  = pstart;
   pthreadnew->Parameter = pparam;
   stackpointer          = stackpointer - sizeof(PDWORD);
   *(PDWORD)stackpointer = pthreadnew;
                 
   /* Klone ourself */   
   __MultiThreading++;
   tid= __mt_clone(&__ThreadNew, CLONE_VM | CLONE_FS | CLONE_FILES | 
                   CLONE_SIGHAND | SIGCHLD, newstack);   
 
   if(tid<0)
      /* error */
      
   /* wait until the new thread is ready initializing */             
   pthread=imalloc(sizeof(THREAD_OBJECT));   
   if(pthread==NULL)
      /* error */
      
   pthread->Kernel.Magic       = MAGIC_THREAD;
   pthread->Kernel.Class       = &__ThreadClass;   
   pthread->Kernel.HandleCount = 2;
   pthread->Kernel.Security    = psa;
   pthread->Thread.ThreadId    = tid | LINUX_THREADID_MASK;
   hthread=__HandleAlloc(pthread, GENERIC_ALL); 
      
   THREAD_LEAVE;
   return hthread;
}


int __mt_clone(void (*fn)(void *), int flags, void *newstack)
{
   /*
   * Do clone() system call. We need to do the low-level stuff
   * entirely in assembly as we're returning with a different
   * stack in the child process and we couldn't otherwise guarantee
   * that the program doesn't use the old stack incorrectly.
   *
   * Parameters to clone() system call:
   *	%eax - __NR_clone, clone system call number
   *	%ebx - clone_flags, bitmap of cloned data
   *	%ecx - new stack pointer for cloned child
   *
   * In this example %ebx is CLONE_VM | CLONE_FS | CLONE_FILES |
   * CLONE_SIGHAND which shares as much as possible between parent
   * and child. (We or in the signal to be sent on child termination
   * into clone_flags: SIGCHLD makes the cloned process work like
   * a "normal" unix child process)
   *
   * The clone() system call returns (in %eax) the pid of the newly
   * cloned process to the parent, and 0 to the cloned process. If
   * an error occurs, the return value will be the negative errno.
   *
   * In the child process, we will do a "jsr" to the requested function
   * and then do a "exit()" system call which will terminate the child.
   */
   __asm__ __volatile__(
      "int $0x80\n\t"		/* Linux/i386 system call */
      "testl %0,%0\n\t"		/* check return value */
      "jne 1f\n\t"		/* jump if parent */
      "call *%3\n\t"		/* start subthread function */
      "movl %2,%0\n\t"
      "int $0x80\n"		/* exit system call: exit subthread */
      "1:\t"
      :"=a" (retval)
      :"0" (__NR_clone),"i" (__NR_exit),
       "r" (fn),
       "b" (flags)
       "c" (newstack));
		 
   return retval;
}

int mt_getpid(void)
{
   __asm__ __volatile__
   (
      "int $0x80\n\t"		/* Linux/i386 system call */
      :"=a" (retval)
      :"0" (__NR_getpid)
   );		 
   return retval;
}

void __ThreadNew(PTHREAD_NEW pnew)
{
   __ThreadInitTeb();
   __TebStackHigh=pnew->StackHigh;
   __TebStackLow=pnew->StackLow;
   __ModuleInitTls(*****);
   __ModuleInitThread(****);

   retval=pnew->Function(pnew->Parameter);  

   __ThreadExit(retval);
}
#endif