/*
 * kernel/virtual.c
 * Copyright (C) 1996, Onno Hovers, All rights reserved
 *
 * Win32 virtual memory functions
 * the amount of work that has gone into getting this as close 
 * to Win32 as possible was really huge
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; see the file COPYING.LIB. If
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 * Cambridge, MA 02139, USA.  
 */
 
 
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <win32/defines.h>
#include <win32/error.h>
#include <win32/kernel.h>
#include "imalloc.h"
#include "error.h"
#include "wfc.h"

#include "virtual.h"

static PVM_AREA FirstArea = 0;


/******************************************************************
*                         __VirtualInit                           *
*                                                                 *
* Initialize virtual memory management structures                 *
*                                                                 *
*******************************************************************/

BOOL __VirtualInit(VOID)
{   
  PVM_AREA   area;
  PVM_REGION region;
  LPVOID     start;
  LPVOID     end;
 
  __DevZero=open("/dev/zero",O_RDONLY, 0 );
  
#ifdef linux
#ifdef __ELF__
  start=(LPVOID) 0x8000000;
#else
  start=NULL;
#endif
#endif
  end  =sbrk(0) + 0x800000;
   
  area			= imalloc( sizeof(VM_AREA) );
  region		= imalloc( sizeof(VM_REGION) );
  region->Base		= start;
  region->Size		= start-end;
  region->State		= PAGE_EXECUTE_READ | MEM_COMMIT;
  region->Next		= NULL;
  area->Base		= start;
  area->Size		= start-end;
  area->First		= region;
  area->Protect		= PAGE_EXECUTE_READ;
  area->Type		= VM_INTERN;  
  area->Descriptor	= -1;
  area->Offset		= 0;
  __VirtualInsertArea(area);
   
  area			= imalloc( sizeof(VM_AREA) );
  region		= imalloc( sizeof(VM_REGION) );
  region->Base		= (LPVOID) 0x40000000;
  region->Size		= 0x01000000;
  region->State		= PAGE_EXECUTE_READ | MEM_COMMIT;
  region->Next		= NULL;
  area->Base		= (LPVOID) 0x40000000;
  area->Size		= 0x01000000;
  area->First		= region;
  area->Protect		= PAGE_EXECUTE_READ;
  area->Type		= VM_INTERN;
  area->Descriptor	= -1;
  area->Offset		= 0;
  __VirtualInsertArea(area);
   
  return TRUE;
}

/*********************************************************************
*                             __VirtualDump                          *
*                                                                    *
* Dump all areas for debugging purpose                               *
*                                                                    *
*********************************************************************/

VOID __VirtualDump(VOID)
{
  PVM_AREA 	area;
  PVM_REGION	region;
  
  area=FirstArea;

  dprintf("Dump of the virtual address space\n");  
  if(!area)
    dprintf("__VirtualDump: FirstArea is 0\n");

  while(area)
  {
     dprintf("[[0x%08lX-0x%08lX]]\t"
            "prot=0x%lX\ttype=0x%lX\tdesc=0x%lX\toffset=0x%lX\n",
            (ULONG) area->Base,
            (ULONG) (area->Base+area->Size),
            area->Protect, 
            area->Type, 
            area->Descriptor, 
            area->Offset );

     region=area->First;
     while(region)
     {
         dprintf("[ 0x%08lX-0x%08lX ]\tprot=0x%lX\n",
                (ULONG) region->Base,
                (ULONG) (region->Base+region->Size),
                region->State);
                    
        region=region->Next;
     }         
     area=area->Next;
  };
}

/********** VM_AREA and VM_REGION administration functions ***********/

/*********************************************************************
*                              __VirtualFindArea                     *
*                                                                    *
* Find an area by address                                            *
*********************************************************************/

PVM_AREA __VirtualFindArea(PCVOID addr)
{
   PVM_AREA area;
      
   area=FirstArea;
   while(area)
   {
      if(area->Base > addr)
         return NULL;
      if(addr<area->Base+area->Size)
         return area;     
      area=area->Next;
   }
   return NULL;
}

/*********************************************************************
*                              __VirtualFindRegion                   *
*                                                                    *
* Find a region by address                                           *
*********************************************************************/

PVM_REGION __VirtualFindRegion(PVM_AREA area, PCVOID addr)
{
   PVM_REGION region;
     
   region=area->First;
   while(region)
   {
      if(region->Base > addr)
         return NULL;
      if(addr<region->Base+region->Size)
         return region;     
      region=region->Next;
   }
   return NULL;
}

/*********************************************************************
*                           __VirtualIsFree                          *
*                                                                    *
* Check if a range of memory is free                                 *
*********************************************************************/

BOOL __VirtualIsFree(PCVOID start, DWORD size)
{
   PCVOID end;
   PVM_AREA area;
   
   if(!size)
      return TRUE;
      
   /* First see if this area is not reserved in advance */
   if((start < (LPVOID) 0x00001000) || (start+size > (LPVOID) 0xC0000000))
      return FALSE;
      
   /* Then, check our lists*/
   end=start+size;
   
   area=FirstArea;
   while(area)
   {   
      if( (end > area->Base) && ( start <= area->Base+area->Size) )
         return FALSE;
       
      area=area->Next;
   } 
   return TRUE;
}   

/*********************************************************************
*                           __VirtualFindFree                        *
*                                                                    *
* Finds a free range of memory                                       *
**********************************************************************/

PVOID __VirtualFindFree(ULONG size, BOOL ftopdown, BOOL fhigh)
{
   PVM_AREA area;
   PVOID    addr;
   PVOID    baseaddr;
   PVOID    endaddr;
   PVOID    lastaddr;
   
   if(!size)
      return 0;
   
   if(!fhigh)
   {     
      if (WIN95_ISMS)   
      {
         /* 
          * the address space between 0x0001000 and 0x00400000 is for old 
          * DOS and 16 bit Windows in Win95 
          */
         baseaddr=(LPVOID) 0x00400000;
         endaddr =(LPVOID) 0x80000000;
      }
      else if (WIN32S_ISMS)
      {
         /*
          *  Win32s only knows a shared address space between
          *  0x80000000 and 0xC0000000
          */
         baseaddr=(LPVOID) 0x80000000;
         endaddr =(LPVOID) 0xC0000000;
      }
      else
      {
         baseaddr=(LPVOID) 0x00010000;
         endaddr =(LPVOID) 0x7FFF0000;
      }            
      addr=(LPVOID) 0x00010000;
   }
   else
   {   
      if (NT_ISMS)  /* NT does not know any high area for shared stuff */
      {
         baseaddr=(LPVOID) 0x00010000;
         endaddr =(LPVOID) 0x7FFF0000;
      }
      else
      {
         baseaddr=(LPVOID) 0x80000000;
         endaddr =(LPVOID) 0xC0000000;
      } 
   }
    
   /* we have the first area, now find a free region of memory */
   area=FirstArea;
   lastaddr=NULL;
   addr=baseaddr;
         
   while((area)&&(area->Base<addr))
      area=area->Next;
   
   if(!area)
   {
      if(!ftopdown)
         return baseaddr;
      else
         return endaddr-ROUNDUP(size,0x00010000);
   }
  
   while(area) 
   {
      if( (LPVOID) ROUNDDOWN((ULONG)
                   (area->Base),0x10000)-addr > size)
      {        
         if((addr>=baseaddr) && (addr+size<=endaddr))
         {
            if(!ftopdown)
               return addr;
            else
               lastaddr=addr;
         }               
      }
      addr=(LPVOID) ROUNDUP((ULONG)
                   (area->Base+area->Size),0x10000);
      area=area->Next;
             
   }     
   if((addr>=baseaddr) && (addr+size<=endaddr))        
      return addr;
                  
   if(!lastaddr) 
   {
      dprintf("__VirtualFindFree: could not find a free range, size=0x%lX\n",
               (ULONG) addr);
      return 0;
   }
   return lastaddr;
}


/********************************************************************
*                            __VirtualFreeArea                      *
*                                                                   *
* Release an area                                                   *
*                                                                   *
********************************************************************/

BOOL __VirtualFreeArea(PVM_AREA freearea)
{ 
   PVM_AREA    prevarea;
   PVM_AREA    area;
   PVM_REGION  region;
   PVM_REGION  nextregion;
    
   area=FirstArea;
   
   if(!area)
   {
      dprintf("__VirtualFree: FirstArea is 0, should not happen\n");
      return FALSE;
   }
      
   /* find our area */
   prevarea=NULL;
   while(area != freearea) 
   {
      prevarea= area;
      area=     area->Next;   
      if(!area) 
         return FALSE; 
   }
   
   region=freearea->First;
   /* now free all our regions */   
   while(region)
   {        
      nextregion = region->Next;   
      ifree(region);
      region=nextregion;
   }
   
   /* update the linked list */
   if(prevarea)   
      prevarea->Next=freearea->Next;
   else
      FirstArea=freearea->Next;

   ifree(freearea);
         
   return TRUE;
}

/********************************************************************
*                            __VirtualInsertArea                    *
*                                                                   *
* Insert an area                                                    *
*                                                                   *
********************************************************************/

BOOL __VirtualInsertArea( PVM_AREA parea )
{
   LPVOID   start;
   ULONG    size;
   PVM_AREA search;
   	
   start=parea->Base;
   size= parea->Size;   	
   /* insert our area into the list */
   if(!FirstArea)
   {	    
      parea->Next=NULL;
      FirstArea=parea;      
   }
   else if(start<FirstArea->Base)
   {
      parea->Next=FirstArea;
      FirstArea=parea;
   }
   else
   {	
      search=FirstArea;
      while(search->Next)
      {	   
         if(search->Next->Base>=start+size)
            break;   
         search=search->Next;
      }  
      parea->Next  =search->Next;
      search->Next=parea;
   }          	  
   return TRUE;
}	

/********************************************************************
*                            __VirtualSetState                      *
*                                                                   *
* Set the state of a range of memory                                *
********************************************************************/

BOOL __VirtualSetState( PVM_AREA area, LPVOID newstart, ULONG newsize, 
                 ULONG state, PULONG poldstate, BOOL fprotect)
{        
   PVM_REGION prevregion;
   PVM_REGION nextregion;  
   PVM_REGION newregion;
   PVM_REGION region;
   PVM_REGION region2;
   
   PVOID      oldstart;
   PVOID      oldend; 
   ULONG      oldsize;
   ULONG      oldstate;
   PVOID      newend;

   
   /* get the area */ 
   if(!area)
      area=__VirtualFindArea(newstart);
      
   if(!area)
      return __ErrorReturnFalse(ERROR_INVALID_ADDRESS); 
      
   /* is this area immutable ? */
   if(area->Type == VM_INTERN)
      { dprintf("__VirtualSetState: internal area\n"); return FALSE; }
 
   newregion        = imalloc(sizeof(VM_REGION));
   newregion->Base  = newstart;
   newregion->Size  = newsize;
   newregion->State = state;
      
   newend     = newstart+newsize;     
   prevregion = NULL;   
   region     = area->First;
      
   if(!region)
   {
      area->First=newregion;
   }
   else
   {
      /* change the rest of the regions */  
      while(region->Base < newend)
      {           
         /* get the dimensions of the old region */
         oldstart	= region->Base;
         oldsize	= region->Size;
         oldstate	= region->State;
         state		= state | (oldstate & MEM_TYPE_MASK);
         oldend		= oldstart+oldsize; 
         nextregion	= region->Next;      
         
         if((oldend>=newstart)&&(newend>=oldstart)&&
            ((!fprotect)||(state!=MEM_FREE)))
         {
            /* merge and free */           
            if(region->State==state)
            {
               if(oldstart<newstart)               
                  newstart=oldstart;
               if(oldend>newend)
                  newend=oldend;        
                   
               if(prevregion)       
                  prevregion->Next=newregion;
               else
                  area->First=newregion;
                  
               newregion->Base=newstart;
               newregion->Size=newend-newstart;
               newregion->Next=region->Next;
               ifree(region);                            
            }
            /* just free it */
            else if((newstart<=oldstart)&&(oldend<=newend))
            {                               
               if(prevregion)       
                  prevregion->Next=newregion;
               else
                  area->First=newregion;
               newregion->Next=region->Next;
               ifree(region);            
               
               __VirtualChangeState(area, oldstart, oldsize, oldstate, state);
            }
            /* split one time */
            else if((oldstart<newstart)&&(newstart<oldend)&&(oldend<=newend))
            {
               /* shrink the region */
               region->Size	= newstart-oldstart;             
               newregion->Next	= region->Next;
               region->Next	= newregion;
               prevregion	= region;
               
               __VirtualChangeState(area, newstart, oldend-newstart, oldstate, state);                                    
            }            
            else if((newstart<=oldstart)&&(oldstart<newend)&&(newend<oldend))
            {
               /* shrink the region */
               region->Base	= newend;
               region->Size	= oldend-newend;

               if(prevregion)       
                  prevregion->Next=newregion;
               else
                  area->First=newregion;               

               newregion->Next	= region; 
               
               __VirtualChangeState(area, oldstart, newend-oldstart, oldstate, state); 
               break;  
            }
            /* split two times */
            else if((oldstart<newstart)&&(newend<oldend))
            {
               /* split the area in three parts */
               region2		= imalloc(sizeof(VM_REGION));          
               memcpy(region2,region,sizeof(VM_REGION)); 
          
               region2->Base	= newend;
               region2->Size	= oldend-newend;
               region->Size	= newstart-oldstart;

               region2->Next	= region->Next;                    
               newregion->Next	= region2;       
               region->Next	= newregion;
               
               __VirtualChangeState(area, newstart, newsize, oldstate, state );
               break;
            }
            else
               prevregion=region;               
         }
         else
            prevregion=region;
            
         region=nextregion;
      
         if(!region)
            return TRUE;
      }
   } 
   return TRUE;
}

/* unix functions */
/*********************************************************************
*                         __VirtualWinProtToUnix                     *
*                                                                    *
* Translates the windows protection to a unix style protection for   *
* mmap                                                               *
**********************************************************************/

ULONG __VirtualWinProtToUnix(DWORD winprot)
{
   int prot;
  
   winprot = winprot & (~(PAGE_NOCACHE));

   switch(winprot)
   {
      case PAGE_READONLY:
         prot=PROT_READ;
         break;
      case PAGE_READWRITE:
         prot=PROT_READ|PROT_WRITE;
         break;
      case PAGE_WRITECOPY:
         prot=PROT_WRITE;
         break;
      case PAGE_EXECUTE:
         prot=PROT_EXEC;
         break;
      case PAGE_EXECUTE_READ:
         prot=PROT_EXEC|PROT_READ;
         break;
      case PAGE_EXECUTE_READWRITE:
         prot=PROT_EXEC|PROT_READ|PROT_WRITE;
         break;
      case PAGE_EXECUTE_WRITECOPY:
         prot=PROT_EXEC|PROT_WRITE;
         break;
      case PAGE_NOACCESS:
      default:
         prot=PROT_NONE;
         break;
   }
   return prot;
}

/********************************************************************
*                         __VirtualChangeState                      *
*                                                                   *
* changes the state of a region of memory                           *
********************************************************************/
BOOL __VirtualChangeState(PVM_AREA area, LPVOID base, ULONG size,
                          ULONG oldstate, ULONG newstate)
{   
#ifdef NOT
    dprintf("__VirtualChangeState(area= 0x%lX, base=0x%lX, size=0x%lX,"
            " oldstate=0x%lX, newstate=0x%lX\n",
            (ULONG) area, (ULONG) base, size, oldstate, newstate );
    
    __VirtualDump();            
#endif

    if((area->Type==VM_IMAGE)||(area->Type==VM_INTERN))
       return TRUE;
    
    if((oldstate&MEM_ALLOC_MASK)==(newstate&MEM_ALLOC_MASK))
    {
       if((oldstate&MEM_PROTECT_MASK)==(newstate&MEM_PROTECT_MASK))
          return TRUE;
                    
       if(mprotect(base, size, __VirtualWinProtToUnix(newstate&MEM_PROTECT_MASK))
          ==-1)
          return __ErrorReturnFalse(__ErrorFromErrno(errno));                  
    }
    else if(newstate&MEM_FREE)
    {
       if(munmap(base, size)==-1)
          return __ErrorReturnFalse(__ErrorFromErrno(errno));
    }
    else
    {
       if(mmap(base, size, __VirtualWinProtToUnix(newstate&MEM_PROTECT_MASK), area->Flags,
               area->Descriptor, area->Offset)==(void*) -1)
          return __ErrorReturnFalse(__ErrorFromErrno(errno));
    }    
    return TRUE;
}


#ifdef NOTYET
/********************************************************************
*                            __VirtualMapping                       *
*                                                                   *
* Create an area in order to map in a file, or to get memory        *
********************************************************************/

PVOID __VirtualMapping( PVOID start, ULONG size, ULONG winprot, 
                 PMAPPING_OBJECT pmapping, ULONGLONG offset )
{
   PVOID 	retval;
   PVM_AREA	area;
   PVM_REGION	region;
   ULONG 	prot;
   ULONG 	flags;   

   /* sanity checks */
   if(!size)
      return 0;
   
   /* find an/our area */
   if(start)
   {
#ifdef VM_NO_SECRETS   
      if(!__VirtualIsFree(start, size))
         return __ErrorReturnNull(ERROR_NOACCESS);    
#else
      return __ErrorReturnNull(ERROR_NOACCESS);
#endif
      /* note that most Unices choose another address when the one 
       * provided is already in use 
       */         
#ifndef VM_NO_SECRETS
     
#endif
   }
   else
   {
#ifdef VM_NO_SECRETS   
      start=__VirtualFindFree(size);
      if(!start)
         return __ErrorReturnFalse(ERROR_OUTOFMEMORY);
#else
      /* warning: very dirty hack!! */
      start=mmap(NULL, size + 0x10000 - PAGE_SIZE, PROT_NONE, 
                 MAP_ANON | MAP_PRIVATE, -1, 0);
      start=ROUNDUP(start, 0x10000);                 
      munmap(start,size + 0x10000 - PAGE_SIZE);
#endif
   }
   	
   /* allocate our area and region */
   area  =imalloc(sizeof(VM_AREA));
   if(!area)
      return NULL;
   region=imalloc(sizeof(VM_REGION));
   if(!region)
      { ifree(area); return NULL; }
      
   /* map our area */
   switch(how)
   {  
      case VM_MEMORY:
         if(!(winflags & MEM_COMMIT))
            prot = PROT_NONE;
         else
            prot = __VirtualWinProtToUnix(winprot);
            
      	 retval = mmap((void *) start, size, prot, 
      	               MAP_ANON | MAP_PRIVATE, -1, 0);
      	 break;      	 
      case VM_FILE:
         prot  = __VirtualAccessToUnix(winprot);
         retval = mmap((void *) start, size, prot, flags, fd, offset);
         break;
      case VM_SHARED:
         flags=   __VirtualAccessToShared(winprot);
         retval=  shmat(fd, start, flags);
         break;
      case VM_IMAGE:
         /* images are really mapped by the loader */
         retval = mmap((void *) start, size, prot, 
                       MAP_ANON | MAP_PRIVATE, -1, 0);
         break;
      case VM_INTERN:
         /* internal mappings are reserved internally */
         retval= start;
         break;  
      case VM_STACK:
         /* mapping a stack */
         retval= mmap(NULL, size, PROT_READ| PROT_WRITE, 
                      MAP_ANON | MAP_PRIVATE | MAP_GROWSDOWN, -1, 0);
         break;
   }
   /* fill in our region */
   region->Next  = NULL;
   region->Base	 = start;
   region->Size  = size;
   region->State = winprot | winflags; 
   
   /* fill in our area */        
   area->Base		= start;  
   area->Size		= size;
   area->Protect	= winprot;
   area->Type		= how;
   area->Descriptor	= fd;
   area->Offset		= offset;   
   area->First		= region;
 
   /* insert our area into the list */
   __VirtualInsertArea(area);
   return start;
}	
#endif

/* functions for internal use */
/********************************************************************
*                        __VirtualReserve                           *
*                                                                   *
* Reserve private memory                                            *
********************************************************************/

PVOID __VirtualReserve(PVOID start, ULONG size, ULONG winflags)                      
{
   PVM_AREA	area;
   PVM_REGION	region;
   ULONG 	prot;

   /* sanity checks */
   if(!size)
      return 0;
   
   /* find an/our area */
   if(start)
   {
#ifdef VM_NO_SECRETS   
      if(!__VirtualIsFree(start, size))
         return __ErrorReturnNull(ERROR_NOACCESS);    
#else
      return __ErrorReturnNull(ERROR_NOACCESS);
#endif          
   }
   else
   {
#ifdef VM_NO_SECRETS   
      start=__VirtualFindFree(size, winflags & MEM_TOP_DOWN, 
                              winflags & MEM_HIGH);
      if(!start)
         return __ErrorReturnNull(ERROR_OUTOFMEMORY);
#else
      /* warning: very dirty hack!! */
      start=mmap(NULL, size + 0x10000 - PAGE_SIZE, PROT_NONE, 
                 MAP_ANON | MAP_PRIVATE, -1, 0);
      start=ROUNDUP(start, 0x10000);                 
      munmap(start,size + 0x10000 - PAGE_SIZE);
#endif
   }
   	
   /* allocate our area and region */
   area  =imalloc(sizeof(VM_AREA));
   if(!area)
      return NULL;
   region=imalloc(sizeof(VM_REGION));
   if(!region)
      { ifree(area); return NULL; }
      
   /* map our area */
   if(!(winflags & MEM_COMMIT))
      prot = PROT_NONE;
   else
      prot = __VirtualWinProtToUnix(winflags & MEM_PROTECT_MASK);

#ifdef VM_REALLY_RESERVE
   /* reserve/commit memory */
   if(start != mmap((void *) start, size, prot, 
                 MAP_ANON | MAP_PRIVATE | MAP_FIXED, __DevZero, 0))
      return __ErrorReturnNull(__ErrorFromErrno(errno));
#else
   /* commit memory */
   if(winflags & MEM_COMMIT)
      if(start != mmap((void *) start, size, prot, 
                     MAP_ANON | MAP_PRIVATE | MAP_FIXED, __DevZero, 0))
          return __ErrorReturnNull(__ErrorFromErrno(errno));
#endif

   /* fill in our region */
   region->Next  = NULL;
   region->Base	 = start;
   region->Size  = size;
   region->State = winflags | MEM_PRIVATE; 
   
   /* fill in our area */        
   area->Base		= start;  
   area->Size		= size;
   area->Protect	= winflags & MEM_PROTECT_MASK;
   area->Type		= VM_MEMORY;
   area->Flags		= MAP_ANON | MAP_PRIVATE; 
   area->Descriptor	= 0;
   area->Offset		= 0;   
   area->First		= region;
 
   /* insert our area into the list */
   __VirtualInsertArea(area);
        
   return start;
}

/********************************************************************
*                        __VirtualAllocStack                        *
* Allocate a stack                                                  *
********************************************************************/
LPVOID __VirtualAllocStack(ULONG size)
{
   PVM_AREA	area;
   PVM_REGION	region;
   LPVOID 	start;
   ULONG 	prot;

   /* sanity checks */
   if(!size)
      return 0;
   
#ifdef VM_NO_SECRETS   
   start=__VirtualFindFree(size, MEM_TOP_DOWN, 0 );
   if(!start)
      return __ErrorReturnNull(ERROR_OUTOFMEMORY);
#else
   /* warning: very dirty hack!! */
   start=mmap(NULL, size + 0x10000 - PAGE_SIZE, PROT_NONE, 
              MAP_ANON | MAP_PRIVATE, -1, 0);
   munmap(start,size + 0x10000 - PAGE_SIZE);                
   start=ROUNDUP(start, 0x10000);                 
#endif  
   	
   /* allocate our area and region */
   area  =imalloc(sizeof(VM_AREA));
   if(!area)
      return NULL;
   region=imalloc(sizeof(VM_REGION));
   if(!region)
      { ifree(area); return NULL; }
      
   /* map our area */  
   prot = PROT_READ | PROT_WRITE | PROT_EXEC ;

   start = mmap((void *) start, size, prot, 
                 MAP_ANON | MAP_PRIVATE | MAP_GROWSDOWN, __DevZero, 0);

   /* fill in our region */
   region->Next  = NULL;
   region->Base	 = start;
   region->Size  = size;
   region->State = PAGE_EXECUTE_READWRITE | MEM_COMMIT | MEM_PRIVATE ;
   
   /* fill in our area */        
   area->Base		= start;  
   area->Size		= size;
   area->Protect	= PAGE_EXECUTE_READWRITE;
   area->Type		= VM_STACK;
   area->Flags		= MAP_ANON | MAP_PRIVATE | MAP_GROWSDOWN; 
   area->Descriptor	= 0;
   area->Offset		= 0;   
   area->First		= region;
 
   /* insert our area into the list */
   __VirtualInsertArea(area);
   return start;
}

/********************************************************************
*                     __VirtualAllocImage                           *
*                                                                   *
* Allocate space for mapping an EXE or DLL                          *
********************************************************************/

PVOID __VirtualAllocImage(LPVOID start, ULONG size, BOOL failifreloc)
{
   PVM_AREA	area;
   PVM_REGION	region;
   ULONG 	prot;

   /* sanity checks */
   if(!size)
      return 0;
   
   /* find an/our area */
   if(start)
   {
#ifdef VM_NO_SECRETS   
      if(!__VirtualIsFree(start, size)&&(failifreloc))
         return __ErrorReturnNull(ERROR_NOACCESS);    
#else
      if(failifreloc)
         return __ErrorReturnNull(ERROR_NOACCESS);
#endif          
   }

#ifdef VM_NO_SECRETS   
   start=__VirtualFindFree(size, 0, 0);
   if(!start)
       return __ErrorReturnNull(ERROR_OUTOFMEMORY);
#else
   /* warning: very dirty hack!! */
   start=mmap(NULL, size + 0x10000 - PAGE_SIZE, PROT_NONE, 
              MAP_ANON | MAP_PRIVATE, -1, 0);
   munmap(start,size + 0x10000 - PAGE_SIZE);              
   start=ROUNDUP(start, 0x10000);                  
#endif
     	
   /* allocate our area and region */
   area  =imalloc(sizeof(VM_AREA));
   if(!area)
      return NULL;
   region=imalloc(sizeof(VM_REGION));
   if(!region)
      { ifree(area); return NULL; }
      
   /* map our area */
   prot = PROT_NONE;

#ifdef VM_REALLY_RESERVE            
   start = mmap((void *) start, size, prot, 
                 MAP_ANON | MAP_PRIVATE, __DevZero, 0);
#endif

   /* fill in our region */
   region->Next  = NULL;
   region->Base	 = start;
   region->Size  = size;
   region->State = PAGE_NOACCESS | MEM_RESERVE | MEM_IMAGE; 
   
   /* fill in our area */        
   area->Base		= start;  
   area->Size		= size;
   area->Protect	= PAGE_NOACCESS;
   area->Type		= VM_IMAGE;
   area->Flags		= 0;
   area->Descriptor	= 0;
   area->Offset		= 0;   
   area->First		= region;
 
   /* insert our area into the list */
   __VirtualInsertArea(area);
   return start;
}

/********************************************************************
*                        __VirtualCommit                            *
*                                                                   *
* Commit region of memory within an area                            *
********************************************************************/

BOOL __VirtualCommit(PVOID start, ULONG size, ULONG protect)
{
   return __VirtualSetState(NULL, start, size, 
                     (protect & MEM_PROTECT_MASK) | MEM_COMMIT,
                     NULL, FALSE);   
}

/********************************************************************
*                        __VirtualDecommit                          *
*                                                                   *
* Decommit region of memory within an area                          *
********************************************************************/

BOOL __VirtualDecommit(PVOID start, ULONG size)
{   
   return __VirtualSetState(NULL, start, size, MEM_RESERVE, NULL, FALSE);
}

/********************************************************************
*                              __VirtualRelease                     *
*                                                                   *
* Free an area of memory                                            *
********************************************************************/

BOOL __VirtualRelease(PVOID start)
{   
   PVM_AREA	area;
   
   /* find the area */
   area=__VirtualFindArea(start);
   if(!area) 
   {
      dprintf("__VirtualRelease: cannot find area\n");  
      return __ErrorReturnFalse( ERROR_INVALID_PARAMETER );
   }
         
   /* is this area immutable ? */
   if(area->Type == VM_INTERN)
      { dprintf("__VirtualRelease: internal area\n"); return FALSE; }
         
   munmap(area->Base, area->Size);
   __VirtualFreeArea(area);
   return TRUE;
}



/* win32 functions */
/********************************************************************
*                      VirtualAlloc - KERNEL32                      *
*                                                                   *
* Allocates or commits memory                                       *
********************************************************************/


LPVOID WINAPI VirtualAlloc(LPVOID address, DWORD size,
                           DWORD alloctype, DWORD protect)
{    
   LPVOID end;

   aprintf("VirtualAlloc( 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n", 
           (ULONG) address, size, alloctype, protect);
           
   /* round addresses to the page size */
   end=    (LPVOID) ROUNDUP((DWORD) address+(DWORD) size, 0x1000) ;
   address=(LPVOID) ROUNDDOWN((DWORD) address, 0x1000);
   size=end-address;      
   
   /* TASM32 4.0 HACK */ 
   if((alloctype & MEM_RESERVE)||(address==NULL))
   {
      dprintf("Reserving address 0x%lX, size 0x%lX\n",
              (ULONG) address, size);
                           
      address= __VirtualReserve( address, size, alloctype | protect );                    
   }
   if(alloctype & MEM_COMMIT)
   {
      dprintf("Commiting address 0x%lX, size 0x%lX, protect 0x%lX\n",
              (ULONG) address, size, protect);
               
      if(!__VirtualCommit( address, size, protect))
         address=NULL;
   }
   return address;
}

/********************************************************************
*                   VirtualFree - KERNEL32                          *
*                                                                   *
*  Frees or decommits memory                                        *
********************************************************************/
BOOL WINAPI VirtualFree(LPVOID address, DWORD size, DWORD freetype)
{
   LPVOID end;
   BOOL   retval;
                
   aprintf("VirtualFree( 0x%lX, 0x%lX, 0x%lX )\n", 
           (ULONG) address, size, freetype);    
   /* round addresses to the page size */
   end=    (LPVOID) ROUNDUP((DWORD) address+ size, 0x1000) ;
   address=(LPVOID) ROUNDDOWN((DWORD) address, 0x1000);     
 
   if(freetype & MEM_RELEASE)
   {
      dprintf("Releasing address 0x%lX \n", (ULONG) address);
      retval = __VirtualRelease(address);
   }
   else
   {
      size=end-address;
      dprintf("Decommiting address 0x%lX, size 0x%lX \n", 
              (ULONG) address, size);                          
      retval=__VirtualDecommit(address, size);         
   }   
   return retval;                           
}

/********************************************************************
*                   VirtualProtect - KERNEL32                       *
*                                                                   *
*  Changes the protection of a range of memory                      *
********************************************************************/

BOOL WINAPI VirtualProtect(LPVOID address, DWORD size, DWORD winprot,
                    LPDWORD poldprot)
{
   aprintf("VirtualProtect( 0x%lX, 0x%lX, 0x%lX, 0x%lX )\n", 
           (ULONG) address, size, winprot, (ULONG) poldprot);
   return __VirtualSetState(NULL, address, size, winprot & MEM_PROTECT_MASK,
                     poldprot, TRUE);   
}

/********************************************************************
*                   VirtualQuery - KERNEL32                         *
*                                                                   *
*  Queries memory at a specified address                            *
********************************************************************/

DWORD WINAPI VirtualQuery(LPCVOID address, PMEMORY_BASIC_INFORMATION pbuf, 
                  DWORD size )
{
   PVM_AREA   area;
   PVM_REGION region;

   aprintf("VirtualQuery( 0x%lX, 0x%lX, 0x%lX )\n", 
           (ULONG) address, (ULONG) pbuf, size);
           
   if(size<sizeof(MEMORY_BASIC_INFORMATION))
     { SetLastError(ERROR_INVALID_PARAMETER); return 0; }         

   memset(pbuf,0,size);
   pbuf->BaseAddress = (LPVOID) address;

   /* find our area */
   area=FirstArea;
   while(address<area->Base+area->Size)
   {      
      area=area->Next;
      if(!area)
      {
         if((!NT_ISMS)&&(address >= (LPVOID) 0x80000000 ))      
            pbuf->RegionSize = (LPVOID) 0xC0000000 - address;
         else            
            pbuf->RegionSize = (LPVOID) 0x80000000 - address;
         pbuf->State= MEM_FREE;
         return sizeof(MEMORY_BASIC_INFORMATION);
      }                    
   }
   if(area->Base > address)
   {
      pbuf->RegionSize=area->Base-address;
      pbuf->State= MEM_FREE;
      return sizeof(MEMORY_BASIC_INFORMATION);      
   }
          
   pbuf->AllocationBase    = area->Base;
   pbuf->AllocationProtect = area->Protect;

   region=__VirtualFindRegion(area, address);
   if(!region)
   {
      SetLastError(ERROR_INVALID_PARAMETER);
      return 0;         
   } 
   pbuf->RegionSize  = region->Size - (address - region->Base);
   pbuf->State       = region->State & MEM_STATE_MASK;
   pbuf->Protect     = region->State & MEM_PROTECT_MASK;
   pbuf->Type        = region->State & MEM_TYPE_MASK;         
   return sizeof(MEMORY_BASIC_INFORMATION);  
} 